# Copyright 2019 Shift Cryptosecurity AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Library to interact with a BitBox02 device. """

from __future__ import print_function
import sys

__version__ = "6.2.0"

if sys.version_info.major != 3 or sys.version_info.minor < 6:
    print(
        "Python version is {}.{}, but 3.6+ is required by this script.".format(
            sys.version_info.major, sys.version_info.minor
        ),
        file=sys.stderr,
    )
    sys.exit(1)

try:
    import hid

    hid.device  # pylint: disable=pointless-statement
except AttributeError:
    print(
        "Unable to reference hid.device; maybe hid package is masking "
        "hidapi? Try:\n\t$ pip3 uninstall hid",
        file=sys.stderr,
    )
    sys.exit(1)

# pylint: disable=wrong-import-position
from .bitbox02 import (
    btc_sign_needs_prevtxs,
    Backup,
    BitBox02,
    BTCInputType,
    BTCOutputExternal,
    BTCOutputInternal,
    BTCOutputType,
    BTCPrevTxInputType,
    BTCPrevTxOutputType,
    DuplicateEntryException,
    hww,
    btc,
    cardano,
    common,
    eth,
    system,
)
from .bootloader import Bootloader



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class CheckBackupRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SILENT_FIELD_NUMBER: builtins.int
    silent: builtins.bool
    def __init__(self,
        *,
        silent: builtins.bool = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["silent",b"silent"]) -> None: ...
global___CheckBackupRequest = CheckBackupRequest

class CheckBackupResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    id: typing.Text
    def __init__(self,
        *,
        id: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id"]) -> None: ...
global___CheckBackupResponse = CheckBackupResponse

class CreateBackupRequest(google.protobuf.message.Message):
    """Timestamp must be in UTC"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    TIMESTAMP_FIELD_NUMBER: builtins.int
    TIMEZONE_OFFSET_FIELD_NUMBER: builtins.int
    timestamp: builtins.int
    timezone_offset: builtins.int
    def __init__(self,
        *,
        timestamp: builtins.int = ...,
        timezone_offset: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["timestamp",b"timestamp","timezone_offset",b"timezone_offset"]) -> None: ...
global___CreateBackupRequest = CreateBackupRequest

class ListBackupsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___ListBackupsRequest = ListBackupsRequest

class BackupInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    TIMESTAMP_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    id: typing.Text
    timestamp: builtins.int
    name: typing.Text
    """uint32 timezone_offset = 3;"""

    def __init__(self,
        *,
        id: typing.Text = ...,
        timestamp: builtins.int = ...,
        name: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id","name",b"name","timestamp",b"timestamp"]) -> None: ...
global___BackupInfo = BackupInfo

class ListBackupsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    INFO_FIELD_NUMBER: builtins.int
    @property
    def info(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___BackupInfo]: ...
    def __init__(self,
        *,
        info: typing.Optional[typing.Iterable[global___BackupInfo]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["info",b"info"]) -> None: ...
global___ListBackupsResponse = ListBackupsResponse

class RestoreBackupRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    TIMESTAMP_FIELD_NUMBER: builtins.int
    TIMEZONE_OFFSET_FIELD_NUMBER: builtins.int
    id: typing.Text
    timestamp: builtins.int
    timezone_offset: builtins.int
    def __init__(self,
        *,
        id: typing.Text = ...,
        timestamp: builtins.int = ...,
        timezone_offset: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id","timestamp",b"timestamp","timezone_offset",b"timezone_offset"]) -> None: ...
global___RestoreBackupRequest = RestoreBackupRequest



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
from . import backup_commands_pb2
from . import bitbox02_system_pb2
from . import btc_pb2
import builtins
from . import cardano_pb2
from . import common_pb2
from . import eth_pb2
import google.protobuf.descriptor
import google.protobuf.message
from . import keystore_pb2
from . import mnemonic_pb2
from . import perform_attestation_pb2
from . import system_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Error(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    CODE_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    code: builtins.int
    message: typing.Text
    def __init__(self,
        *,
        code: builtins.int = ...,
        message: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["code",b"code","message",b"message"]) -> None: ...
global___Error = Error

class Success(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___Success = Success

class Request(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    DEVICE_NAME_FIELD_NUMBER: builtins.int
    DEVICE_LANGUAGE_FIELD_NUMBER: builtins.int
    DEVICE_INFO_FIELD_NUMBER: builtins.int
    SET_PASSWORD_FIELD_NUMBER: builtins.int
    CREATE_BACKUP_FIELD_NUMBER: builtins.int
    SHOW_MNEMONIC_FIELD_NUMBER: builtins.int
    BTC_PUB_FIELD_NUMBER: builtins.int
    BTC_SIGN_INIT_FIELD_NUMBER: builtins.int
    BTC_SIGN_INPUT_FIELD_NUMBER: builtins.int
    BTC_SIGN_OUTPUT_FIELD_NUMBER: builtins.int
    INSERT_REMOVE_SDCARD_FIELD_NUMBER: builtins.int
    CHECK_SDCARD_FIELD_NUMBER: builtins.int
    SET_MNEMONIC_PASSPHRASE_ENABLED_FIELD_NUMBER: builtins.int
    LIST_BACKUPS_FIELD_NUMBER: builtins.int
    RESTORE_BACKUP_FIELD_NUMBER: builtins.int
    PERFORM_ATTESTATION_FIELD_NUMBER: builtins.int
    REBOOT_FIELD_NUMBER: builtins.int
    CHECK_BACKUP_FIELD_NUMBER: builtins.int
    ETH_FIELD_NUMBER: builtins.int
    RESET_FIELD_NUMBER: builtins.int
    RESTORE_FROM_MNEMONIC_FIELD_NUMBER: builtins.int
    FINGERPRINT_FIELD_NUMBER: builtins.int
    BTC_FIELD_NUMBER: builtins.int
    ELECTRUM_ENCRYPTION_KEY_FIELD_NUMBER: builtins.int
    CARDANO_FIELD_NUMBER: builtins.int
    @property
    def device_name(self) -> bitbox02_system_pb2.SetDeviceNameRequest:
        """removed: RandomNumberRequest random_number = 1;"""
        pass
    @property
    def device_language(self) -> bitbox02_system_pb2.SetDeviceLanguageRequest: ...
    @property
    def device_info(self) -> bitbox02_system_pb2.DeviceInfoRequest: ...
    @property
    def set_password(self) -> bitbox02_system_pb2.SetPasswordRequest: ...
    @property
    def create_backup(self) -> backup_commands_pb2.CreateBackupRequest: ...
    @property
    def show_mnemonic(self) -> mnemonic_pb2.ShowMnemonicRequest: ...
    @property
    def btc_pub(self) -> btc_pb2.BTCPubRequest: ...
    @property
    def btc_sign_init(self) -> btc_pb2.BTCSignInitRequest: ...
    @property
    def btc_sign_input(self) -> btc_pb2.BTCSignInputRequest: ...
    @property
    def btc_sign_output(self) -> btc_pb2.BTCSignOutputRequest: ...
    @property
    def insert_remove_sdcard(self) -> bitbox02_system_pb2.InsertRemoveSDCardRequest: ...
    @property
    def check_sdcard(self) -> bitbox02_system_pb2.CheckSDCardRequest: ...
    @property
    def set_mnemonic_passphrase_enabled(self) -> mnemonic_pb2.SetMnemonicPassphraseEnabledRequest: ...
    @property
    def list_backups(self) -> backup_commands_pb2.ListBackupsRequest: ...
    @property
    def restore_backup(self) -> backup_commands_pb2.RestoreBackupRequest: ...
    @property
    def perform_attestation(self) -> perform_attestation_pb2.PerformAttestationRequest: ...
    @property
    def reboot(self) -> system_pb2.RebootRequest: ...
    @property
    def check_backup(self) -> backup_commands_pb2.CheckBackupRequest: ...
    @property
    def eth(self) -> eth_pb2.ETHRequest: ...
    @property
    def reset(self) -> bitbox02_system_pb2.ResetRequest: ...
    @property
    def restore_from_mnemonic(self) -> mnemonic_pb2.RestoreFromMnemonicRequest: ...
    @property
    def fingerprint(self) -> common_pb2.RootFingerprintRequest:
        """removed: BitBoxBaseRequest bitboxbase = 23;"""
        pass
    @property
    def btc(self) -> btc_pb2.BTCRequest: ...
    @property
    def electrum_encryption_key(self) -> keystore_pb2.ElectrumEncryptionKeyRequest: ...
    @property
    def cardano(self) -> cardano_pb2.CardanoRequest: ...
    def __init__(self,
        *,
        device_name: typing.Optional[bitbox02_system_pb2.SetDeviceNameRequest] = ...,
        device_language: typing.Optional[bitbox02_system_pb2.SetDeviceLanguageRequest] = ...,
        device_info: typing.Optional[bitbox02_system_pb2.DeviceInfoRequest] = ...,
        set_password: typing.Optional[bitbox02_system_pb2.SetPasswordRequest] = ...,
        create_backup: typing.Optional[backup_commands_pb2.CreateBackupRequest] = ...,
        show_mnemonic: typing.Optional[mnemonic_pb2.ShowMnemonicRequest] = ...,
        btc_pub: typing.Optional[btc_pb2.BTCPubRequest] = ...,
        btc_sign_init: typing.Optional[btc_pb2.BTCSignInitRequest] = ...,
        btc_sign_input: typing.Optional[btc_pb2.BTCSignInputRequest] = ...,
        btc_sign_output: typing.Optional[btc_pb2.BTCSignOutputRequest] = ...,
        insert_remove_sdcard: typing.Optional[bitbox02_system_pb2.InsertRemoveSDCardRequest] = ...,
        check_sdcard: typing.Optional[bitbox02_system_pb2.CheckSDCardRequest] = ...,
        set_mnemonic_passphrase_enabled: typing.Optional[mnemonic_pb2.SetMnemonicPassphraseEnabledRequest] = ...,
        list_backups: typing.Optional[backup_commands_pb2.ListBackupsRequest] = ...,
        restore_backup: typing.Optional[backup_commands_pb2.RestoreBackupRequest] = ...,
        perform_attestation: typing.Optional[perform_attestation_pb2.PerformAttestationRequest] = ...,
        reboot: typing.Optional[system_pb2.RebootRequest] = ...,
        check_backup: typing.Optional[backup_commands_pb2.CheckBackupRequest] = ...,
        eth: typing.Optional[eth_pb2.ETHRequest] = ...,
        reset: typing.Optional[bitbox02_system_pb2.ResetRequest] = ...,
        restore_from_mnemonic: typing.Optional[mnemonic_pb2.RestoreFromMnemonicRequest] = ...,
        fingerprint: typing.Optional[common_pb2.RootFingerprintRequest] = ...,
        btc: typing.Optional[btc_pb2.BTCRequest] = ...,
        electrum_encryption_key: typing.Optional[keystore_pb2.ElectrumEncryptionKeyRequest] = ...,
        cardano: typing.Optional[cardano_pb2.CardanoRequest] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["btc",b"btc","btc_pub",b"btc_pub","btc_sign_init",b"btc_sign_init","btc_sign_input",b"btc_sign_input","btc_sign_output",b"btc_sign_output","cardano",b"cardano","check_backup",b"check_backup","check_sdcard",b"check_sdcard","create_backup",b"create_backup","device_info",b"device_info","device_language",b"device_language","device_name",b"device_name","electrum_encryption_key",b"electrum_encryption_key","eth",b"eth","fingerprint",b"fingerprint","insert_remove_sdcard",b"insert_remove_sdcard","list_backups",b"list_backups","perform_attestation",b"perform_attestation","reboot",b"reboot","request",b"request","reset",b"reset","restore_backup",b"restore_backup","restore_from_mnemonic",b"restore_from_mnemonic","set_mnemonic_passphrase_enabled",b"set_mnemonic_passphrase_enabled","set_password",b"set_password","show_mnemonic",b"show_mnemonic"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["btc",b"btc","btc_pub",b"btc_pub","btc_sign_init",b"btc_sign_init","btc_sign_input",b"btc_sign_input","btc_sign_output",b"btc_sign_output","cardano",b"cardano","check_backup",b"check_backup","check_sdcard",b"check_sdcard","create_backup",b"create_backup","device_info",b"device_info","device_language",b"device_language","device_name",b"device_name","electrum_encryption_key",b"electrum_encryption_key","eth",b"eth","fingerprint",b"fingerprint","insert_remove_sdcard",b"insert_remove_sdcard","list_backups",b"list_backups","perform_attestation",b"perform_attestation","reboot",b"reboot","request",b"request","reset",b"reset","restore_backup",b"restore_backup","restore_from_mnemonic",b"restore_from_mnemonic","set_mnemonic_passphrase_enabled",b"set_mnemonic_passphrase_enabled","set_password",b"set_password","show_mnemonic",b"show_mnemonic"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["request",b"request"]) -> typing.Optional[typing_extensions.Literal["device_name","device_language","device_info","set_password","create_backup","show_mnemonic","btc_pub","btc_sign_init","btc_sign_input","btc_sign_output","insert_remove_sdcard","check_sdcard","set_mnemonic_passphrase_enabled","list_backups","restore_backup","perform_attestation","reboot","check_backup","eth","reset","restore_from_mnemonic","fingerprint","btc","electrum_encryption_key","cardano"]]: ...
global___Request = Request

class Response(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SUCCESS_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    DEVICE_INFO_FIELD_NUMBER: builtins.int
    PUB_FIELD_NUMBER: builtins.int
    BTC_SIGN_NEXT_FIELD_NUMBER: builtins.int
    LIST_BACKUPS_FIELD_NUMBER: builtins.int
    CHECK_BACKUP_FIELD_NUMBER: builtins.int
    PERFORM_ATTESTATION_FIELD_NUMBER: builtins.int
    CHECK_SDCARD_FIELD_NUMBER: builtins.int
    ETH_FIELD_NUMBER: builtins.int
    FINGERPRINT_FIELD_NUMBER: builtins.int
    BTC_FIELD_NUMBER: builtins.int
    ELECTRUM_ENCRYPTION_KEY_FIELD_NUMBER: builtins.int
    CARDANO_FIELD_NUMBER: builtins.int
    @property
    def success(self) -> global___Success: ...
    @property
    def error(self) -> global___Error: ...
    @property
    def device_info(self) -> bitbox02_system_pb2.DeviceInfoResponse:
        """removed: RandomNumberResponse random_number = 3;"""
        pass
    @property
    def pub(self) -> common_pb2.PubResponse: ...
    @property
    def btc_sign_next(self) -> btc_pb2.BTCSignNextResponse: ...
    @property
    def list_backups(self) -> backup_commands_pb2.ListBackupsResponse: ...
    @property
    def check_backup(self) -> backup_commands_pb2.CheckBackupResponse: ...
    @property
    def perform_attestation(self) -> perform_attestation_pb2.PerformAttestationResponse: ...
    @property
    def check_sdcard(self) -> bitbox02_system_pb2.CheckSDCardResponse: ...
    @property
    def eth(self) -> eth_pb2.ETHResponse: ...
    @property
    def fingerprint(self) -> common_pb2.RootFingerprintResponse: ...
    @property
    def btc(self) -> btc_pb2.BTCResponse: ...
    @property
    def electrum_encryption_key(self) -> keystore_pb2.ElectrumEncryptionKeyResponse: ...
    @property
    def cardano(self) -> cardano_pb2.CardanoResponse: ...
    def __init__(self,
        *,
        success: typing.Optional[global___Success] = ...,
        error: typing.Optional[global___Error] = ...,
        device_info: typing.Optional[bitbox02_system_pb2.DeviceInfoResponse] = ...,
        pub: typing.Optional[common_pb2.PubResponse] = ...,
        btc_sign_next: typing.Optional[btc_pb2.BTCSignNextResponse] = ...,
        list_backups: typing.Optional[backup_commands_pb2.ListBackupsResponse] = ...,
        check_backup: typing.Optional[backup_commands_pb2.CheckBackupResponse] = ...,
        perform_attestation: typing.Optional[perform_attestation_pb2.PerformAttestationResponse] = ...,
        check_sdcard: typing.Optional[bitbox02_system_pb2.CheckSDCardResponse] = ...,
        eth: typing.Optional[eth_pb2.ETHResponse] = ...,
        fingerprint: typing.Optional[common_pb2.RootFingerprintResponse] = ...,
        btc: typing.Optional[btc_pb2.BTCResponse] = ...,
        electrum_encryption_key: typing.Optional[keystore_pb2.ElectrumEncryptionKeyResponse] = ...,
        cardano: typing.Optional[cardano_pb2.CardanoResponse] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["btc",b"btc","btc_sign_next",b"btc_sign_next","cardano",b"cardano","check_backup",b"check_backup","check_sdcard",b"check_sdcard","device_info",b"device_info","electrum_encryption_key",b"electrum_encryption_key","error",b"error","eth",b"eth","fingerprint",b"fingerprint","list_backups",b"list_backups","perform_attestation",b"perform_attestation","pub",b"pub","response",b"response","success",b"success"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["btc",b"btc","btc_sign_next",b"btc_sign_next","cardano",b"cardano","check_backup",b"check_backup","check_sdcard",b"check_sdcard","device_info",b"device_info","electrum_encryption_key",b"electrum_encryption_key","error",b"error","eth",b"eth","fingerprint",b"fingerprint","list_backups",b"list_backups","perform_attestation",b"perform_attestation","pub",b"pub","response",b"response","success",b"success"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["response",b"response"]) -> typing.Optional[typing_extensions.Literal["success","error","device_info","pub","btc_sign_next","list_backups","check_backup","perform_attestation","check_sdcard","eth","fingerprint","btc","electrum_encryption_key","cardano"]]: ...
global___Response = Response



"""
Ledger Devices
**************
"""

from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from ..descriptor import (
    MultisigDescriptor,
    PubkeyProvider,
)
from ..hwwclient import HardwareWalletClient
from ..errors import (
    ActionCanceledError,
    BadArgumentError,
    DeviceConnectionError,
    DeviceFailureError,
    UnavailableActionError,
    UnknownDeviceError,
    common_err_msgs,
    handle_errors,
)
from ..common import (
    AddressType,
    Chain,
)
from .ledger_bitcoin.client import (
    createClient,
    NewClient,
    LegacyClient,
    TransportClient,
)
from .ledger_bitcoin.client_base import ApduException
from .ledger_bitcoin.exception import NotSupportedError
from .ledger_bitcoin.wallet import (
    MultisigWallet,
    WalletPolicy,
)
from .ledger_bitcoin.btchip.btchipException import BTChipException

import builtins
import copy
import hid

from ..key import (
    ExtendedKey,
    get_bip44_purpose,
    get_bip44_chain,
    H_,
    is_standard_path,
    KeyOriginInfo,
    parse_path,
)
from .._script import (
    is_p2sh,
    is_p2wsh,
    is_witness,
    parse_multisig,
)
from ..psbt import PSBT
import logging
import re

SIMULATOR_PATH = 'tcp:127.0.0.1:9999'

LEDGER_VENDOR_ID = 0x2c97
LEDGER_MODEL_IDS = {
    0x10: "ledger_nano_s",
    0x40: "ledger_nano_x",
    0x50: "ledger_nano_s_plus",
    0x60: "ledger_stax",
    0x70: "ledger_flex"
}
LEDGER_LEGACY_PRODUCT_IDS = {
    0x0001: "ledger_nano_s",
    0x0004: "ledger_nano_x"
}

# minimal checking of string keypath
def check_keypath(key_path: str) -> bool:
    parts = re.split("/", key_path)
    if parts[0] != "m":
        return False
    # strip hardening chars
    for index in parts[1:]:
        index_int = re.sub('[hH\']', '', index)
        if not index_int.isdigit():
            return False
        if int(index_int) > 0x80000000:
            return False
    return True

bad_args = [
    0x6700, # BTCHIP_SW_INCORRECT_LENGTH
    0x6A80, # BTCHIP_SW_INCORRECT_DATA
    0x6B00, # BTCHIP_SW_INCORRECT_P1_P2
    0x6D00, # BTCHIP_SW_INS_NOT_SUPPORTED
]

cancels = [
    0x6982, # BTCHIP_SW_SECURITY_STATUS_NOT_SATISFIED
    0x6985, # BTCHIP_SW_CONDITIONS_OF_USE_NOT_SATISFIED
]

# The priority of address types we want for signing.
# We want to do Taproot first, then segwit, then legacy
# Higher number is lower priority so that sort does not require reversing.
signing_priority = {
    AddressType.TAP: 0,
    AddressType.WIT: 1,
    AddressType.SH_WIT: 2,
    AddressType.LEGACY: 3,
}

def handle_chip_exception(e: Union[BTChipException, ApduException], func_name: str) -> bool:
    if e.sw in bad_args:
        raise BadArgumentError('Bad argument')
    elif e.sw == 0x6F00: # BTCHIP_SW_TECHNICAL_PROBLEM
        raise DeviceFailureError(e.message)
    elif e.sw == 0x6FAA: # BTCHIP_SW_HALTED
        raise DeviceConnectionError('Device is asleep')
    elif e.sw in cancels:
        raise ActionCanceledError('{} canceled'.format(func_name))
    else:
        raise e

def ledger_exception(f: Callable[..., Any]) -> Any:
    @wraps(f)
    def func(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except ValueError as e:
            raise BadArgumentError(str(e))
        except BTChipException as e:
            handle_chip_exception(e, f.__name__)
        except ApduException as e:
            e.message = e.data.decode("utf-8")
            handle_chip_exception(e, f.__name__)
    return func

# This class extends the HardwareWalletClient for Ledger Nano S and Nano X specific things
class LedgerClient(HardwareWalletClient):

    def __init__(self, path: str, password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> None:
        super(LedgerClient, self).__init__(path, password, expert, chain)

        is_debug = logging.getLogger().getEffectiveLevel() == logging.DEBUG

        try:
            if path.startswith('tcp'):
                split_path = path.split(':')
                server = split_path[1]
                port = int(split_path[2])
                self.transport_client = TransportClient(interface="tcp", server=server, port=port, debug=is_debug)
            else:
                self.transport_client = TransportClient(interface="hid", debug=is_debug, hid_path=path.encode())

            self.client = createClient(self.transport_client, chain=self.chain, debug=is_debug)
        except NotSupportedError as e:
            raise DeviceConnectionError(e.args[2])

    @ledger_exception
    def get_master_fingerprint(self) -> bytes:
        return self.client.get_master_fingerprint()

    @ledger_exception
    def get_pubkey_at_path(self, path: str) -> ExtendedKey:
        path = path.replace("h", "'")
        path = path.replace("H", "'")
        try:
            xpub_str = self.client.get_extended_pubkey(path=path, display=False)
        except NotSupportedError:
            # We will get not supported for non-standard paths
            # If so, try again but with display=True
            xpub_str = self.client.get_extended_pubkey(path=path, display=True)
        return ExtendedKey.deserialize(xpub_str)

    @ledger_exception
    def sign_tx(self, tx: PSBT) -> PSBT:
        """
        Sign a transaction with a Ledger device. Not all transactions can be signed by a Ledger.

        The scripts supported depend on the version of the Bitcoin Application installed on the Ledger.

        For application versions 1.x and 2.0.x:

        - Transactions containing both segwit and non-segwit inputs are not entirely supported; only the segwit inputs will be signed in this case.

        For application versions 2.1.x and above:

        - Only keys derived with standard BIP 44, 49, 84, and 86 derivation paths are supported for single signature addresses.
        """
        master_fp = self.get_master_fingerprint()

        def legacy_sign_tx() -> PSBT:
            client = self.client
            if not isinstance(client, LegacyClient):
                client = LegacyClient(self.transport_client, self.chain)
            wallet = WalletPolicy("", "wpkh(@0/**)", [""])
            legacy_input_sigs = client.sign_psbt(tx, wallet, None)

            for idx, pubkey, sig in legacy_input_sigs:
                psbt_in = tx.inputs[idx]
                psbt_in.partial_sigs[pubkey] = sig
            return tx

        if isinstance(self.client, LegacyClient):
            return legacy_sign_tx()

        # Make a deepcopy of this psbt. We will need to modify it to get signing to work,
        # which will affect the caller's detection for whether signing occured.
        psbt2 = copy.deepcopy(tx)
        if tx.version != 2:
            psbt2.convert_to_v2()

        # Figure out which wallets are signing
        wallets: Dict[bytes, Tuple[int, AddressType, WalletPolicy, Optional[bytes]]] = {}
        pubkeys: Dict[int, bytes] = {}
        for input_num, psbt_in in builtins.enumerate(psbt2.inputs):
            utxo = None
            scriptcode = b""
            if psbt_in.witness_utxo:
                utxo = psbt_in.witness_utxo
            if psbt_in.non_witness_utxo:
                if psbt_in.prev_txid != psbt_in.non_witness_utxo.hash:
                    raise BadArgumentError(f"Input {input_num} has a non_witness_utxo with the wrong hash")
                assert psbt_in.prev_out is not None
                utxo = psbt_in.non_witness_utxo.vout[psbt_in.prev_out]
                psbt_in.witness_utxo = utxo # Make sure that all inputs have witness_utxo too as signing mixed will fail
            if utxo is None:
                continue
            scriptcode = utxo.scriptPubKey

            p2sh = False
            if is_p2sh(scriptcode):
                if len(psbt_in.redeem_script) == 0:
                    continue
                scriptcode = psbt_in.redeem_script
                p2sh = True

            is_wit, wit_ver, _ = is_witness(scriptcode)

            script_addrtype = AddressType.LEGACY
            if is_wit:
                if p2sh:
                    if wit_ver == 0:
                        script_addrtype = AddressType.SH_WIT
                    else:
                        raise BadArgumentError("Cannot have witness v1+ in p2sh")
                else:
                    if wit_ver == 0:
                        script_addrtype = AddressType.WIT
                    elif wit_ver == 1:
                        script_addrtype = AddressType.TAP
                    else:
                        continue

            # Check if P2WSH
            if is_p2wsh(scriptcode):
                if len(psbt_in.witness_script) == 0:
                    continue
                scriptcode = psbt_in.witness_script

            multisig = parse_multisig(scriptcode)
            if multisig is not None:
                k, ms_pubkeys = multisig

                # Figure out the parent xpubs
                key_exprs: List[str] = []
                ok = True
                our_keys = 0
                for pub in ms_pubkeys:
                    if pub in psbt_in.hd_keypaths:
                        pk_origin = psbt_in.hd_keypaths[pub]
                        if pk_origin.fingerprint == master_fp:
                            our_keys += 1
                            pubkeys[input_num] = pub
                        for xpub_bytes, xpub_origin in psbt2.xpub.items():
                            xpub = ExtendedKey.from_bytes(xpub_bytes)
                            if (xpub_origin.fingerprint == pk_origin.fingerprint) and (xpub_origin.path == pk_origin.path[:len(xpub_origin.path)]):
                                key_exprs.append(PubkeyProvider(xpub_origin, xpub.to_string(), None).to_string(hardened_char="'"))
                                break
                        else:
                            # No xpub, Ledger will not accept this multisig
                            ok = False

                if not ok:
                    continue

                # Make and register the MultisigWallet
                msw = MultisigWallet(f"{k} of {len(key_exprs)} Multisig", script_addrtype, k, key_exprs)
                msw_id = msw.id
                if msw_id not in wallets:
                    _, registered_hmac = self.client.register_wallet(msw)
                    wallets[msw_id] = (
                        signing_priority[script_addrtype],
                        script_addrtype,
                        msw,
                        registered_hmac,
                    )
            else:
                def process_origin(origin: KeyOriginInfo) -> None:
                    if not is_standard_path(origin.path, script_addrtype, self.chain):
                        # TODO: Deal with non-default wallets
                        return
                    policy = self._get_singlesig_default_wallet_policy(script_addrtype, origin.path[2])
                    wallets[policy.id] = (
                        signing_priority[script_addrtype],
                        script_addrtype,
                        self._get_singlesig_default_wallet_policy(script_addrtype, origin.path[2]),
                        None, # Wallet hmac
                    )

                for key, origin in psbt_in.hd_keypaths.items():
                    if origin.fingerprint == master_fp:
                        if not multisig:
                            process_origin(origin)
                        pubkeys[input_num] = key

                for key, (leaf_hashes, origin) in psbt_in.tap_bip32_paths.items():
                    # TODO: Support script path signing
                    if key == psbt_in.tap_internal_key and origin.fingerprint == master_fp:
                        process_origin(origin)
                        pubkeys[input_num] = key

        # For each wallet, sign
        for _, (_, addrtype, wallet, wallet_hmac) in sorted(wallets.items(), key=lambda y: y[1][0]):
            if addrtype == AddressType.LEGACY:
                # We need to remove witness_utxo for legacy inputs when signing with legacy otherwise signing will fail
                for psbt_in in psbt2.inputs:
                    utxo = None
                    if psbt_in.witness_utxo:
                        utxo = psbt_in.witness_utxo
                    if utxo is None:
                        continue
                    is_wit, _, _ = is_witness(utxo.scriptPubKey)
                    if not is_wit:
                        psbt_in.witness_utxo = None

            input_sigs = self.client.sign_psbt(psbt2, wallet, wallet_hmac)

            for idx, pubkey, sig in input_sigs:
                psbt_in = psbt2.inputs[idx]

                utxo = None
                if psbt_in.witness_utxo:
                    utxo = psbt_in.witness_utxo
                if psbt_in.non_witness_utxo:
                    assert psbt_in.prev_out is not None
                    utxo = psbt_in.non_witness_utxo.vout[psbt_in.prev_out]
                assert utxo is not None

                is_wit, wit_ver, _ = utxo.is_witness()

                if is_wit and wit_ver >= 1:
                    # TODO: Deal with script path signatures
                    # For now, assume key path signature
                    psbt_in.tap_key_sig = sig
                else:
                    psbt_in.partial_sigs[pubkey] = sig

        # Extract the sigs from psbt2 and put them into tx
        for sig_in, psbt_in in zip(psbt2.inputs, tx.inputs):
            psbt_in.partial_sigs.update(sig_in.partial_sigs)
            psbt_in.tap_script_sigs.update(sig_in.tap_script_sigs)
            if len(sig_in.tap_key_sig) != 0 and len(psbt_in.tap_key_sig) == 0:
                psbt_in.tap_key_sig = sig_in.tap_key_sig

        return tx

    @ledger_exception
    def sign_message(self, message: Union[str, bytes], keypath: str) -> str:
        if not check_keypath(keypath):
            raise ValueError("Invalid keypath")

        # Ledger requires the character "'" for hardened derivations since version 2.0.0
        keypath = keypath.replace("h", "'")

        return self.client.sign_message(message, keypath)

    def _get_singlesig_default_wallet_policy(self, addr_type: AddressType, account: int) -> WalletPolicy:
        if addr_type == AddressType.LEGACY:
            template = "pkh(@0/**)"
        elif addr_type == AddressType.WIT:
            template = "wpkh(@0/**)"
        elif addr_type == AddressType.SH_WIT:
            template = "sh(wpkh(@0/**))"
        elif addr_type == AddressType.TAP:
            template = "tr(@0/**)"
        else:
            BadArgumentError("Unknown address type")

        path = [H_(get_bip44_purpose(addr_type)), H_(get_bip44_chain(self.chain)), H_(account)]

        # Build a PubkeyProvider for the key we're going to use
        origin = KeyOriginInfo(self.get_master_fingerprint(), path)
        pk_prov = PubkeyProvider(origin, self.get_pubkey_at_path(f"m{origin._path_string()}").to_string(), None)
        key_str = pk_prov.to_string(hardened_char="'")

        # Make the Wallet object
        return WalletPolicy(name="", descriptor_template=template, keys_info=[key_str])

    @ledger_exception
    def display_singlesig_address(
        self,
        keypath: str,
        addr_type: AddressType,
    ) -> str:
        path = parse_path(keypath)

        if isinstance(self.client, LegacyClient):
            if addr_type == AddressType.LEGACY:
                template = "pkh(@0/**)"
            elif addr_type == AddressType.WIT:
                template = "wpkh(@0/**)"
            elif addr_type == AddressType.SH_WIT:
                template = "sh(wpkh(@0/**))"
            elif addr_type == AddressType.TAP:
                BadArgumentError("Taproot is not supported by this version of the Bitcoin App")
            else:
                BadArgumentError("Unknown address type")

            origin = KeyOriginInfo(self.get_master_fingerprint(), path)
            wallet = WalletPolicy(name="", descriptor_template=template, keys_info=["[{}]".format(origin.to_string(hardened_char="'"))])
        else:
            if not is_standard_path(path, addr_type, self.chain):
                raise BadArgumentError("Ledger requires BIP 44 standard paths")

            wallet = self._get_singlesig_default_wallet_policy(addr_type, path[2])

        return self.client.get_wallet_address(wallet, None, path[-2], path[-1], True)

    @ledger_exception
    def display_multisig_address(
        self,
        addr_type: AddressType,
        multisig: MultisigDescriptor,
    ) -> str:
        if isinstance(self.client, LegacyClient):
            raise BadArgumentError("Displaying multisignature addresses is not supported by this version of the Bitcoin App")

        def is_valid_der_path(path: Optional[str]) -> bool:
            if path is None:
                return False
            path_parts = path.split("/")
            return len(path_parts) == 3 and path_parts[1] in ["0", "1"] and path_parts[2].isdigit() and 0 <= int(path_parts[2]) <= 0x7fffffff

        if any(not is_valid_der_path(pk.deriv_path) for pk in multisig.pubkeys):
            raise BadArgumentError("Ledger Bitcoin app requires derivation paths ending with /0/* or /1/* for multisig")

        if not (all(pk.deriv_path == multisig.pubkeys[0].deriv_path for pk in multisig.pubkeys)):
            raise BadArgumentError("Ledger Bitcoin app requires all derivation paths to end with /0/*, or all with /1/* for multisig")

        if any(pk.origin is not None and len(pk.origin.path) > 4 for pk in multisig.pubkeys):
            raise BadArgumentError("Ledger Bitcoin app requires extended keys with derivation length at most 4")

        def format_key_info(pubkey: PubkeyProvider) -> str:
            assert pubkey.origin is not None and pubkey.extkey is not None
            hardened_char = "'"
            return f"[{pubkey.origin.to_string(hardened_char=hardened_char)}]{pubkey.extkey.to_string()}"

        keys_info = [format_key_info(pk) for pk in multisig.pubkeys]

        multisig_wallet = MultisigWallet(f"{multisig.thresh} of {len(keys_info)} Multisig", addr_type, multisig.thresh, keys_info=keys_info, sorted=multisig.is_sorted)
        _, registered_hmac = self.client.register_wallet(multisig_wallet)

        assert multisig.pubkeys[0].deriv_path is not None  # already checked above with is_valid_der_path
        change = 0 if multisig.pubkeys[0].deriv_path[:3] == "/0/" else 1
        address_index = int(multisig.pubkeys[0].deriv_path.split("/")[2])

        return self.client.get_wallet_address(multisig_wallet, registered_hmac, change, address_index, True)

    def setup_device(self, label: str = "", passphrase: str = "") -> bool:
        """
        Ledgers do not support setup via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Ledger Nano S and X do not support software setup')

    def wipe_device(self) -> bool:
        """
        Ledgers do not support wiping via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Ledger Nano S and X do not support wiping via software')

    def restore_device(self, label: str = "", word_count: int = 24) -> bool:
        """
        Ledgers do not support restoring via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Ledger Nano S and X do not support restoring via software')

    def backup_device(self, label: str = "", passphrase: str = "") -> bool:
        """
        Ledgers do not support backing up via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Ledger Nano S and X do not support creating a backup via software')

    def close(self) -> None:
        self.client.stop()

    def prompt_pin(self) -> bool:
        """
        Ledgers do not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Ledger Nano S and X do not need a PIN sent from the host')

    def send_pin(self, pin: str) -> bool:
        """
        Ledgers do not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Ledger Nano S and X do not need a PIN sent from the host')

    def toggle_passphrase(self) -> bool:
        """
        Ledgers do not support toggling passphrase from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Ledger Nano S and X do not support toggling passphrase from the host')

    @ledger_exception
    def can_sign_taproot(self) -> bool:
        """
        Ledgers support Taproot if the Bitcoin App version greater than 2.0.0; support here is for versions 2.1.0 and above.

        :returns: True if Bitcoin App version is greater than or equal to 2.1.0, and not the "Legacy" release. False otherwise.
        """
        return isinstance(self.client, NewClient)


def enumerate(password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN, allow_emulators: bool = False) -> List[Dict[str, Any]]:
    results = []
    devices = []
    devices.extend(hid.enumerate(LEDGER_VENDOR_ID, 0))
    if allow_emulators:
        devices.append({'path': SIMULATOR_PATH.encode(), 'interface_number': 0, 'product_id': 0x1000})

    for d in devices:
        if ('interface_number' in d and d['interface_number'] == 0
                or ('usage_page' in d and d['usage_page'] == 0xffa0)):
            d_data: Dict[str, Any] = {}

            path = d['path'].decode()
            d_data['type'] = 'ledger'
            model = d['product_id'] >> 8
            if model in LEDGER_MODEL_IDS.keys():
                d_data['model'] = LEDGER_MODEL_IDS[model]
            elif d['product_id'] in LEDGER_LEGACY_PRODUCT_IDS.keys():
                d_data['model'] = LEDGER_LEGACY_PRODUCT_IDS[d['product_id']]
            else:
                continue
            d_data['label'] = None
            d_data['path'] = path

            if path == SIMULATOR_PATH:
                d_data['model'] += '_simulator'

            client = None
            with handle_errors(common_err_msgs["enumerate"], d_data):
                try:
                    client = LedgerClient(path, password)
                    d_data['fingerprint'] = client.get_master_fingerprint().hex()
                    d_data['needs_pin_sent'] = False
                    d_data['needs_passphrase_sent'] = False
                except (BTChipException, ConnectionRefusedError):
                    # Ignore simulator if there's an exception, means it isn't there
                    if path == SIMULATOR_PATH:
                        continue
                    else:
                        raise
                except UnknownDeviceError:
                    # This only happens if the ledger is not in the Bitcoin app, so skip it
                    continue

            if client:
                client.close()

            results.append(d_data)

    return results



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: eth.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2
from . import antiklepto_pb2 as antiklepto__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\teth.proto\x12\x14shiftcrypto.bitbox02\x1a\x0c\x63ommon.proto\x1a\x10\x61ntiklepto.proto\"\xf4\x01\n\rETHPubRequest\x12\x0f\n\x07keypath\x18\x01 \x03(\r\x12+\n\x04\x63oin\x18\x02 \x01(\x0e\x32\x1d.shiftcrypto.bitbox02.ETHCoin\x12\x43\n\x0boutput_type\x18\x03 \x01(\x0e\x32..shiftcrypto.bitbox02.ETHPubRequest.OutputType\x12\x0f\n\x07\x64isplay\x18\x04 \x01(\x08\x12\x18\n\x10\x63ontract_address\x18\x05 \x01(\x0c\x12\x10\n\x08\x63hain_id\x18\x06 \x01(\x04\"#\n\nOutputType\x12\x0b\n\x07\x41\x44\x44RESS\x10\x00\x12\x08\n\x04XPUB\x10\x01\"\x99\x02\n\x0e\x45THSignRequest\x12+\n\x04\x63oin\x18\x01 \x01(\x0e\x32\x1d.shiftcrypto.bitbox02.ETHCoin\x12\x0f\n\x07keypath\x18\x02 \x03(\r\x12\r\n\x05nonce\x18\x03 \x01(\x0c\x12\x11\n\tgas_price\x18\x04 \x01(\x0c\x12\x11\n\tgas_limit\x18\x05 \x01(\x0c\x12\x11\n\trecipient\x18\x06 \x01(\x0c\x12\r\n\x05value\x18\x07 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x08 \x01(\x0c\x12R\n\x15host_nonce_commitment\x18\t \x01(\x0b\x32\x33.shiftcrypto.bitbox02.AntiKleptoHostNonceCommitment\x12\x10\n\x08\x63hain_id\x18\n \x01(\x04\"\xc8\x01\n\x15\x45THSignMessageRequest\x12+\n\x04\x63oin\x18\x01 \x01(\x0e\x32\x1d.shiftcrypto.bitbox02.ETHCoin\x12\x0f\n\x07keypath\x18\x02 \x03(\r\x12\x0b\n\x03msg\x18\x03 \x01(\x0c\x12R\n\x15host_nonce_commitment\x18\x04 \x01(\x0b\x32\x33.shiftcrypto.bitbox02.AntiKleptoHostNonceCommitment\x12\x10\n\x08\x63hain_id\x18\x05 \x01(\x04\"$\n\x0f\x45THSignResponse\x12\x11\n\tsignature\x18\x01 \x01(\x0c\"\xfb\x05\n\x1a\x45THSignTypedMessageRequest\x12\x10\n\x08\x63hain_id\x18\x01 \x01(\x04\x12\x0f\n\x07keypath\x18\x02 \x03(\r\x12J\n\x05types\x18\x03 \x03(\x0b\x32;.shiftcrypto.bitbox02.ETHSignTypedMessageRequest.StructType\x12\x14\n\x0cprimary_type\x18\x04 \x01(\t\x12R\n\x15host_nonce_commitment\x18\x05 \x01(\x0b\x32\x33.shiftcrypto.bitbox02.AntiKleptoHostNonceCommitment\x1a\xc9\x01\n\nMemberType\x12G\n\x04type\x18\x01 \x01(\x0e\x32\x39.shiftcrypto.bitbox02.ETHSignTypedMessageRequest.DataType\x12\x0c\n\x04size\x18\x02 \x01(\r\x12\x13\n\x0bstruct_name\x18\x03 \x01(\t\x12O\n\narray_type\x18\x04 \x01(\x0b\x32;.shiftcrypto.bitbox02.ETHSignTypedMessageRequest.MemberType\x1a\x61\n\x06Member\x12\x0c\n\x04name\x18\x01 \x01(\t\x12I\n\x04type\x18\x02 \x01(\x0b\x32;.shiftcrypto.bitbox02.ETHSignTypedMessageRequest.MemberType\x1a\x64\n\nStructType\x12\x0c\n\x04name\x18\x01 \x01(\t\x12H\n\x07members\x18\x02 \x03(\x0b\x32\x37.shiftcrypto.bitbox02.ETHSignTypedMessageRequest.Member\"o\n\x08\x44\x61taType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05\x42YTES\x10\x01\x12\x08\n\x04UINT\x10\x02\x12\x07\n\x03INT\x10\x03\x12\x08\n\x04\x42OOL\x10\x04\x12\x0b\n\x07\x41\x44\x44RESS\x10\x05\x12\n\n\x06STRING\x10\x06\x12\t\n\x05\x41RRAY\x10\x07\x12\n\n\x06STRUCT\x10\x08\"\xb4\x01\n\x1c\x45THTypedMessageValueResponse\x12R\n\x0broot_object\x18\x01 \x01(\x0e\x32=.shiftcrypto.bitbox02.ETHTypedMessageValueResponse.RootObject\x12\x0c\n\x04path\x18\x02 \x03(\r\"2\n\nRootObject\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x44OMAIN\x10\x01\x12\x0b\n\x07MESSAGE\x10\x02\",\n\x1b\x45THTypedMessageValueRequest\x12\r\n\x05value\x18\x01 \x01(\x0c\"\xae\x03\n\nETHRequest\x12\x32\n\x03pub\x18\x01 \x01(\x0b\x32#.shiftcrypto.bitbox02.ETHPubRequestH\x00\x12\x34\n\x04sign\x18\x02 \x01(\x0b\x32$.shiftcrypto.bitbox02.ETHSignRequestH\x00\x12?\n\x08sign_msg\x18\x03 \x01(\x0b\x32+.shiftcrypto.bitbox02.ETHSignMessageRequestH\x00\x12P\n\x14\x61ntiklepto_signature\x18\x04 \x01(\x0b\x32\x30.shiftcrypto.bitbox02.AntiKleptoSignatureRequestH\x00\x12J\n\x0esign_typed_msg\x18\x05 \x01(\x0b\x32\x30.shiftcrypto.bitbox02.ETHSignTypedMessageRequestH\x00\x12L\n\x0ftyped_msg_value\x18\x06 \x01(\x0b\x32\x31.shiftcrypto.bitbox02.ETHTypedMessageValueRequestH\x00\x42\t\n\x07request\"\xab\x02\n\x0b\x45THResponse\x12\x30\n\x03pub\x18\x01 \x01(\x0b\x32!.shiftcrypto.bitbox02.PubResponseH\x00\x12\x35\n\x04sign\x18\x02 \x01(\x0b\x32%.shiftcrypto.bitbox02.ETHSignResponseH\x00\x12X\n\x1c\x61ntiklepto_signer_commitment\x18\x03 \x01(\x0b\x32\x30.shiftcrypto.bitbox02.AntiKleptoSignerCommitmentH\x00\x12M\n\x0ftyped_msg_value\x18\x04 \x01(\x0b\x32\x32.shiftcrypto.bitbox02.ETHTypedMessageValueResponseH\x00\x42\n\n\x08response*2\n\x07\x45THCoin\x12\x07\n\x03\x45TH\x10\x00\x12\x0e\n\nRopstenETH\x10\x01\x12\x0e\n\nRinkebyETH\x10\x02\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'eth_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ETHCOIN._serialized_start=2569
  _ETHCOIN._serialized_end=2619
  _ETHPUBREQUEST._serialized_start=68
  _ETHPUBREQUEST._serialized_end=312
  _ETHPUBREQUEST_OUTPUTTYPE._serialized_start=277
  _ETHPUBREQUEST_OUTPUTTYPE._serialized_end=312
  _ETHSIGNREQUEST._serialized_start=315
  _ETHSIGNREQUEST._serialized_end=596
  _ETHSIGNMESSAGEREQUEST._serialized_start=599
  _ETHSIGNMESSAGEREQUEST._serialized_end=799
  _ETHSIGNRESPONSE._serialized_start=801
  _ETHSIGNRESPONSE._serialized_end=837
  _ETHSIGNTYPEDMESSAGEREQUEST._serialized_start=840
  _ETHSIGNTYPEDMESSAGEREQUEST._serialized_end=1603
  _ETHSIGNTYPEDMESSAGEREQUEST_MEMBERTYPE._serialized_start=1088
  _ETHSIGNTYPEDMESSAGEREQUEST_MEMBERTYPE._serialized_end=1289
  _ETHSIGNTYPEDMESSAGEREQUEST_MEMBER._serialized_start=1291
  _ETHSIGNTYPEDMESSAGEREQUEST_MEMBER._serialized_end=1388
  _ETHSIGNTYPEDMESSAGEREQUEST_STRUCTTYPE._serialized_start=1390
  _ETHSIGNTYPEDMESSAGEREQUEST_STRUCTTYPE._serialized_end=1490
  _ETHSIGNTYPEDMESSAGEREQUEST_DATATYPE._serialized_start=1492
  _ETHSIGNTYPEDMESSAGEREQUEST_DATATYPE._serialized_end=1603
  _ETHTYPEDMESSAGEVALUERESPONSE._serialized_start=1606
  _ETHTYPEDMESSAGEVALUERESPONSE._serialized_end=1786
  _ETHTYPEDMESSAGEVALUERESPONSE_ROOTOBJECT._serialized_start=1736
  _ETHTYPEDMESSAGEVALUERESPONSE_ROOTOBJECT._serialized_end=1786
  _ETHTYPEDMESSAGEVALUEREQUEST._serialized_start=1788
  _ETHTYPEDMESSAGEVALUEREQUEST._serialized_end=1832
  _ETHREQUEST._serialized_start=1835
  _ETHREQUEST._serialized_end=2265
  _ETHRESPONSE._serialized_start=2268
  _ETHRESPONSE._serialized_end=2567
# @@protoc_insertion_point(module_scope)



# Copyright 2019 Shift Cryptosecurity AG
# Copyright 2020 Shift Crypto AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BitBox02"""

import os
import sys
import time
from datetime import datetime
from typing import Optional, List, Dict, Tuple, Any, Generator, Union, Sequence
from typing_extensions import TypedDict

import semver

from ..communication import (
    BitBoxCommonAPI,
    Bitbox02Exception,
    ERR_GENERIC,
    ERR_DUPLICATE_ENTRY,
)

from .secp256k1 import antiklepto_host_commit, antiklepto_verify

from ..communication.generated import hww_pb2 as hww
from ..communication.generated import eth_pb2 as eth
from ..communication.generated import btc_pb2 as btc
from ..communication.generated import cardano_pb2 as cardano
from ..communication.generated import mnemonic_pb2 as mnemonic
from ..communication.generated import bitbox02_system_pb2 as bitbox02_system
from ..communication.generated import backup_commands_pb2 as backup
from ..communication.generated import common_pb2 as common
from ..communication.generated import keystore_pb2 as keystore
from ..communication.generated import antiklepto_pb2 as antiklepto

# pylint: disable=unused-import
# We export it in __init__.py
from ..communication.generated import system_pb2 as system

try:
    # Optional rlp dependency only needed to sign ethereum transactions.
    # pylint: disable=import-error
    import rlp
except ModuleNotFoundError:
    pass

HARDENED = 0x80000000

Backup = Tuple[str, str, datetime]


class DuplicateEntryException(Exception):
    pass


def is_taproot(script_config: btc.BTCScriptConfigWithKeypath) -> bool:
    # pylint: disable=no-member
    return (
        script_config.script_config.WhichOneof("config") == "simple_type"
        and script_config.script_config.simple_type == btc.BTCScriptConfig.P2TR
    )


def btc_sign_needs_prevtxs(script_configs: Sequence[btc.BTCScriptConfigWithKeypath]) -> bool:
    """Returns True if the prev_tx field in BTCInputType needs to be
    populated before calling btc_sign(). This is the case if there are
    any non-taproot inputs in the transaction to be signed.
    """
    return not all(map(is_taproot, script_configs))


class BTCPrevTxInputType(TypedDict):
    prev_out_hash: bytes
    prev_out_index: int
    signature_script: bytes
    sequence: int


class BTCPrevTxOutputType(TypedDict):
    value: int
    pubkey_script: bytes


class BTCPrevTxType(TypedDict):
    version: int
    locktime: int
    inputs: Sequence[BTCPrevTxInputType]
    outputs: Sequence[BTCPrevTxOutputType]


class BTCInputType(TypedDict):
    prev_out_hash: bytes
    prev_out_index: int
    prev_out_value: int
    sequence: int
    keypath: Sequence[int]
    script_config_index: int
    # Must be the transaction referenced by prev_out_hash. Can be None if `btc_sign_needs_prevtxs()` returns False.
    prev_tx: Optional[BTCPrevTxType]


class BTCOutputInternal:
    # TODO: Use NamedTuple, but not playing well with protobuf types.

    def __init__(self, keypath: Sequence[int], value: int, script_config_index: int):
        """
        keypath: keypath to the change output.
        """
        self.keypath = keypath
        self.value = value
        self.script_config_index = script_config_index


class BTCOutputExternal:
    # TODO: Use NamedTuple, but not playing well with protobuf types.

    def __init__(self, output_type: "btc.BTCOutputType.V", output_payload: bytes, value: int):
        self.type = output_type
        self.payload = output_payload
        self.value = value


BTCOutputType = Union[BTCOutputInternal, BTCOutputExternal]


class BitBox02(BitBoxCommonAPI):
    """Class to communicate with a BitBox02"""

    # pylint: disable=too-many-public-methods

    def device_info(self) -> Dict[str, Any]:
        """
        Returns an object with device information, e.g. name, passphrase status, etc.
        """
        # pylint: disable=no-member
        request = hww.Request()
        device_info_request = bitbox02_system.DeviceInfoRequest()
        request.device_info.CopyFrom(device_info_request)
        response = self._msg_query(request, expected_response="device_info")
        result = {
            "name": response.device_info.name,
            "version": response.device_info.version,
            "initialized": response.device_info.initialized,
            "mnemonic_passphrase_enabled": response.device_info.mnemonic_passphrase_enabled,
            "monotonic_increments_remaining": response.device_info.monotonic_increments_remaining,
        }
        if self.version >= semver.VersionInfo(9, 6, 0):
            result["securechip_model"] = response.device_info.securechip_model

        return result

    def set_device_name(self, device_name: str) -> None:
        # pylint: disable=no-member
        request = hww.Request()
        request.device_name.name = device_name
        self._msg_query(request, expected_response="success")

    def set_password(self, entropy_size: int = 32) -> bool:
        """
        Returns True if the user entered the password correctly (passwords match).
        Returns False otherwise. Entropy size determines the seed size in bytes; must be 16 or 32.
        """
        assert entropy_size in (16, 32)
        if entropy_size == 16:
            self._require_atleast(semver.VersionInfo(9, 6, 0))

        # pylint: disable=no-member
        request = hww.Request()
        request.set_password.entropy = os.urandom(entropy_size)
        try:
            self._msg_query(request, expected_response="success")
        except Bitbox02Exception as err:
            if err.code == ERR_GENERIC:
                return False
            raise
        return True

    def create_backup(self) -> bool:
        """
        Returns True if the backup was created successfully.
        Returns False otherwise.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.create_backup.timestamp = int(time.time())
        request.create_backup.timezone_offset = time.localtime().tm_gmtoff
        try:
            self._msg_query(request, expected_response="success")
        except Bitbox02Exception as err:
            if err.code == ERR_GENERIC:
                return False
            raise
        return True

    def list_backups(self) -> Generator[Backup, None, None]:
        """
        Returns a pair of id and timestamp's strings that identify the backups.
        """
        # pylint: disable=no-member
        self.insert_sdcard()
        request = hww.Request()
        request.list_backups.CopyFrom(backup.ListBackupsRequest())
        response = self._msg_query(request, expected_response="list_backups")
        for info in response.list_backups.info:
            utcdate = datetime.utcfromtimestamp(info.timestamp)
            yield (info.id, info.name, utcdate)

    def restore_backup(self, backup_id: str) -> None:
        """
        Sends a restore API call to the BitBox. Raises a Bitbox02Exception on failure.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.restore_backup.id = backup_id
        request.restore_backup.timestamp = int(time.time())
        request.restore_backup.timezone_offset = time.localtime().tm_gmtoff
        self._msg_query(request, expected_response="success")

    def check_backup(self, silent: bool = False) -> Optional[str]:
        """
        Sends a check backup API call to the BitBox.
        Returns the backup ID if the backup was found and can be restored.
        Otherwise, returns None. If silent is True, the result won't be shown on the device screen.
        """
        # pylint: disable=no-member
        self.insert_sdcard()
        request = hww.Request()
        request.check_backup.CopyFrom(backup.CheckBackupRequest(silent=silent))
        try:
            response = self._msg_query(request, expected_response="check_backup")
        except Bitbox02Exception as err:
            if err.code == ERR_GENERIC:
                return None
            raise
        return response.check_backup.id

    def show_mnemonic(self) -> None:
        """
        Returns True if mnemonic was successfully shown and confirmed.
        Raises a Bitbox02Exception on failure.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.show_mnemonic.CopyFrom(mnemonic.ShowMnemonicRequest())
        self._msg_query(request, expected_response="success")

    def _btc_msg_query(
        self, btc_request: btc.BTCRequest, expected_response: Optional[str] = None
    ) -> btc.BTCResponse:
        """
        Same as _msg_query, but one nesting deeper for bitcoin messages.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.btc.CopyFrom(btc_request)
        btc_response = self._msg_query(request, expected_response="btc").btc
        if (
            expected_response is not None
            and btc_response.WhichOneof("response") != expected_response
        ):
            raise Exception(
                "Unexpected response: {}, expected: {}".format(
                    btc_response.WhichOneof("response"), expected_response
                )
            )
        return btc_response

    def btc_xpub(
        self,
        keypath: Sequence[int],
        coin: "btc.BTCCoin.V" = btc.BTC,
        xpub_type: "btc.BTCPubRequest.XPubType.V" = btc.BTCPubRequest.XPUB,
        display: bool = True,
    ) -> str:
        """
        keypath is a list of child derivation numbers.
        e.g. m/44'/0'/1' corresponds to [44+HARDENED, 0+HARDENED, 1+HARDENED].
        """
        # pylint: disable=no-member,too-many-arguments
        request = hww.Request()
        request.btc_pub.CopyFrom(
            btc.BTCPubRequest(coin=coin, keypath=keypath, xpub_type=xpub_type, display=display)
        )
        return self._msg_query(request).pub.pub

    def btc_address(
        self,
        keypath: Sequence[int],
        coin: "btc.BTCCoin.V" = btc.BTC,
        script_config: btc.BTCScriptConfig = btc.BTCScriptConfig(
            simple_type=btc.BTCScriptConfig.P2WPKH
        ),
        display: bool = True,
    ) -> str:
        """
        keypath is a list of child derivation numbers.
        e.g. m/44'/0'/1'/5/10 corresponds to [44+HARDENED, 0+HARDENED, 1+HARDENED, 5, 10].
        """
        # pylint: disable=no-member,too-many-arguments
        request = hww.Request()
        request.btc_pub.CopyFrom(
            btc.BTCPubRequest(
                coin=coin, keypath=keypath, script_config=script_config, display=display
            )
        )
        return self._msg_query(request).pub.pub

    def btc_is_script_config_registered(
        self, coin: "btc.BTCCoin.V", script_config: btc.BTCScriptConfig, keypath: Sequence[int]
    ) -> bool:
        """
        Returns True if the script config / account is already registered.
        """
        # pylint: disable=no-member
        request = btc.BTCRequest()
        request.is_script_config_registered.CopyFrom(
            btc.BTCIsScriptConfigRegisteredRequest(
                registration=btc.BTCScriptConfigRegistration(
                    coin=coin, script_config=script_config, keypath=keypath
                )
            )
        )
        return self._btc_msg_query(
            request, expected_response="is_script_config_registered"
        ).is_script_config_registered.is_registered

    def btc_register_script_config(
        self,
        coin: "btc.BTCCoin.V",
        script_config: btc.BTCScriptConfig,
        keypath: Sequence[int],
        name: str,
        xpub_type: "btc.BTCRegisterScriptConfigRequest.XPubType.V" = btc.BTCRegisterScriptConfigRequest.XPubType.AUTO_ELECTRUM,
    ) -> None:
        """
        Raises Bitbox02Exception with ERR_USER_ABORT on user abort.
        If name is the empty string, it will be prompted on the device.
        """
        # pylint: disable=no-member,too-many-arguments

        if name == "":
            # prompt on device only available since v9.3.0
            self._require_atleast(semver.VersionInfo(9, 3, 0))

        assert len(name) <= 30

        # pylint: disable=no-member
        request = btc.BTCRequest()
        request.register_script_config.CopyFrom(
            btc.BTCRegisterScriptConfigRequest(
                registration=btc.BTCScriptConfigRegistration(
                    coin=coin, script_config=script_config, keypath=keypath
                ),
                name=name,
                xpub_type=xpub_type,
            )
        )
        try:
            self._btc_msg_query(request, expected_response="success")
        except Bitbox02Exception as err:
            if err.code == ERR_DUPLICATE_ENTRY:
                raise DuplicateEntryException(
                    "A multisig account configuration with this name already exists.\n"
                    "Choose another name."
                )
            raise

    # pylint: disable=too-many-arguments
    def btc_sign(
        self,
        coin: "btc.BTCCoin.V",
        script_configs: Sequence[btc.BTCScriptConfigWithKeypath],
        inputs: Sequence[BTCInputType],
        outputs: Sequence[BTCOutputType],
        version: int = 1,
        locktime: int = 0,
        format_unit: "btc.BTCSignInitRequest.FormatUnit.V" = btc.BTCSignInitRequest.FormatUnit.DEFAULT,
    ) -> Sequence[Tuple[int, bytes]]:
        """
        coin: the first element of all provided keypaths must match the coin:
        - BTC: 0 + HARDENED
        - Testnets: 1 + HARDENED
        - LTC: 2 + HARDENED
        script_configs: types of all inputs and change outputs. The first element of all provided
        keypaths must match this type:
        - SCRIPT_P2PKH: 44 + HARDENED
        - SCRIPT_P2WPKH_P2SH: 49 + HARDENED
        - SCRIPT_P2WPKH: 84 + HARDENED
        - SCRIPT_P2TR: 86 + HARDENED
        inputs: transaction inputs. The previous transactions of the inputs need to be provided
          if `btc_sign_needs_prevtxs()` returns True.
        outputs: transaction outputs. Can be an external output
        (BTCOutputExternal) or an internal output for change (BTCOutputInternal).
        version, locktime: reserved for future use.
        Returns: list of (input index, signature) tuples.
        Raises Bitbox02Exception with ERR_USER_ABORT on user abort.
        """
        # pylint: disable=no-member,too-many-branches,too-many-statements

        # Reserved for future use.
        assert version in (1, 2)

        if any(map(is_taproot, script_configs)):
            self._require_atleast(semver.VersionInfo(9, 10, 0))

        supports_antiklepto = self.version >= semver.VersionInfo(9, 4, 0)

        sigs: List[Tuple[int, bytes]] = []

        # Init request
        request = hww.Request()
        request.btc_sign_init.CopyFrom(
            btc.BTCSignInitRequest(
                coin=coin,
                script_configs=script_configs,
                version=version,
                num_inputs=len(inputs),
                num_outputs=len(outputs),
                locktime=locktime,
                format_unit=format_unit,
            )
        )
        next_response = self._msg_query(request, expected_response="btc_sign_next").btc_sign_next

        is_inputs_pass2 = False
        while True:
            if next_response.type == btc.BTCSignNextResponse.INPUT:
                input_index = next_response.index
                tx_input = inputs[input_index]

                request = hww.Request()
                request.btc_sign_input.CopyFrom(
                    btc.BTCSignInputRequest(
                        prevOutHash=tx_input["prev_out_hash"],
                        prevOutIndex=tx_input["prev_out_index"],
                        prevOutValue=tx_input["prev_out_value"],
                        sequence=tx_input["sequence"],
                        keypath=tx_input["keypath"],
                        script_config_index=tx_input["script_config_index"],
                    )
                )

                # Anti-Klepto protocol not supported yet for Schnorr signatures.
                input_is_schnorr = is_taproot(script_configs[tx_input["script_config_index"]])
                perform_antiklepto = (
                    supports_antiklepto and is_inputs_pass2 and not input_is_schnorr
                )

                if perform_antiklepto:
                    host_nonce = os.urandom(32)
                    request.btc_sign_input.host_nonce_commitment.commitment = (
                        antiklepto_host_commit(host_nonce)
                    )

                next_response = self._msg_query(
                    request, expected_response="btc_sign_next"
                ).btc_sign_next

                if perform_antiklepto:
                    assert next_response.type == btc.BTCSignNextResponse.HOST_NONCE
                    assert next_response.HasField("anti_klepto_signer_commitment")
                    signer_commitment = next_response.anti_klepto_signer_commitment.commitment

                    btc_request = btc.BTCRequest()
                    btc_request.antiklepto_signature.CopyFrom(
                        antiklepto.AntiKleptoSignatureRequest(host_nonce=host_nonce)
                    )
                    next_response = self._btc_msg_query(
                        btc_request, expected_response="sign_next"
                    ).sign_next

                    if self.debug:
                        print(
                            f"For input {input_index}, the host contributed the nonce {host_nonce.hex()}"
                        )

                    assert next_response.has_signature
                    antiklepto_verify(host_nonce, signer_commitment, next_response.signature)

                    if self.debug:
                        print(f"Antiklepto nonce verification PASSED for input {input_index}")

                if is_inputs_pass2:
                    assert next_response.has_signature
                    sigs.append((input_index, next_response.signature))

                if input_index == len(inputs) - 1:
                    is_inputs_pass2 = True

            elif next_response.type == btc.BTCSignNextResponse.PREVTX_INIT:
                prevtx = inputs[next_response.index]["prev_tx"]
                assert prevtx, "Previous transaction missing"
                btc_request = btc.BTCRequest()
                btc_request.prevtx_init.CopyFrom(
                    btc.BTCPrevTxInitRequest(
                        version=prevtx["version"],
                        num_inputs=len(prevtx["inputs"]),
                        num_outputs=len(prevtx["outputs"]),
                        locktime=prevtx["locktime"],
                    )
                )
                next_response = self._btc_msg_query(
                    btc_request, expected_response="sign_next"
                ).sign_next
            elif next_response.type == btc.BTCSignNextResponse.PREVTX_INPUT:
                prevtx = inputs[next_response.index]["prev_tx"]
                assert prevtx, "Previous transaction missing"
                prevtx_input = prevtx["inputs"][next_response.prev_index]
                btc_request = btc.BTCRequest()
                btc_request.prevtx_input.CopyFrom(
                    btc.BTCPrevTxInputRequest(
                        prev_out_hash=prevtx_input["prev_out_hash"],
                        prev_out_index=prevtx_input["prev_out_index"],
                        signature_script=prevtx_input["signature_script"],
                        sequence=prevtx_input["sequence"],
                    )
                )
                next_response = self._btc_msg_query(
                    btc_request, expected_response="sign_next"
                ).sign_next
            elif next_response.type == btc.BTCSignNextResponse.PREVTX_OUTPUT:
                prevtx = inputs[next_response.index]["prev_tx"]
                assert prevtx, "Previous transaction missing"
                prevtx_output = prevtx["outputs"][next_response.prev_index]
                btc_request = btc.BTCRequest()
                btc_request.prevtx_output.CopyFrom(
                    btc.BTCPrevTxOutputRequest(
                        value=prevtx_output["value"], pubkey_script=prevtx_output["pubkey_script"]
                    )
                )
                next_response = self._btc_msg_query(
                    btc_request, expected_response="sign_next"
                ).sign_next
            elif next_response.type == btc.BTCSignNextResponse.OUTPUT:
                output_index = next_response.index
                tx_output = outputs[output_index]

                request = hww.Request()
                if isinstance(tx_output, BTCOutputInternal):
                    request.btc_sign_output.CopyFrom(
                        btc.BTCSignOutputRequest(
                            ours=True,
                            value=tx_output.value,
                            keypath=tx_output.keypath,
                            script_config_index=tx_output.script_config_index,
                        )
                    )
                elif isinstance(tx_output, BTCOutputExternal):
                    request.btc_sign_output.CopyFrom(
                        btc.BTCSignOutputRequest(
                            ours=False,
                            type=tx_output.type,
                            payload=tx_output.payload,
                            value=tx_output.value,
                        )
                    )
                next_response = self._msg_query(
                    request, expected_response="btc_sign_next"
                ).btc_sign_next
            elif next_response.type == btc.BTCSignNextResponse.DONE:
                break
            else:
                raise Exception("unexpected response")
        return sigs

    def btc_sign_msg(
        self, coin: "btc.BTCCoin.V", script_config: btc.BTCScriptConfigWithKeypath, msg: bytes
    ) -> Tuple[bytes, int, bytes]:
        """
        Returns a 64 byte sig, the recoverable id, and a 65 byte signature containing
        the recid, compatible with Electrum.
        """
        # pylint: disable=no-member

        self._require_atleast(semver.VersionInfo(9, 2, 0))

        request = btc.BTCRequest()
        request.sign_message.CopyFrom(
            btc.BTCSignMessageRequest(coin=coin, script_config=script_config, msg=msg)
        )

        supports_antiklepto = self.version >= semver.VersionInfo(9, 5, 0)
        if supports_antiklepto:
            host_nonce = os.urandom(32)

            request.sign_message.host_nonce_commitment.commitment = antiklepto_host_commit(
                host_nonce
            )
            signer_commitment = self._btc_msg_query(
                request, expected_response="antiklepto_signer_commitment"
            ).antiklepto_signer_commitment.commitment

            request = btc.BTCRequest()
            request.antiklepto_signature.CopyFrom(
                antiklepto.AntiKleptoSignatureRequest(host_nonce=host_nonce)
            )

            signature = self._btc_msg_query(
                request, expected_response="sign_message"
            ).sign_message.signature
            antiklepto_verify(host_nonce, signer_commitment, signature[:64])

            if self.debug:
                print("Antiklepto nonce verification PASSED")

        else:
            signature = self._btc_msg_query(
                request, expected_response="sign_message"
            ).sign_message.signature

        sig, recid = signature[:64], signature[64]

        # See https://github.com/spesmilo/electrum/blob/84dc181b6e7bb20e88ef6b98fb8925c5f645a765/electrum/ecc.py#L521-L523
        compressed = 4  # BitBox02 uses only compressed pubkeys
        electrum_sig65 = bytes([27 + compressed + recid]) + sig

        return (sig, recid, electrum_sig65)

    def check_sdcard(self) -> bool:
        # pylint: disable=no-member
        request = hww.Request()
        request.check_sdcard.CopyFrom(bitbox02_system.CheckSDCardRequest())
        response = self._msg_query(request, expected_response="check_sdcard")
        return response.check_sdcard.inserted

    def insert_sdcard(self) -> None:
        # pylint: disable=no-member
        request = hww.Request()
        request.insert_remove_sdcard.CopyFrom(
            bitbox02_system.InsertRemoveSDCardRequest(
                action=bitbox02_system.InsertRemoveSDCardRequest.INSERT_CARD
            )
        )
        self._msg_query(request, expected_response="success")

    def remove_sdcard(self) -> None:
        # pylint: disable=no-member
        request = hww.Request()
        request.insert_remove_sdcard.CopyFrom(
            bitbox02_system.InsertRemoveSDCardRequest(
                action=bitbox02_system.InsertRemoveSDCardRequest.REMOVE_CARD
            )
        )
        self._msg_query(request, expected_response="success")

    def root_fingerprint(self) -> bytes:
        """
        Get the root fingerprint from the bitbox02
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.fingerprint.CopyFrom(common.RootFingerprintRequest())
        response = self._msg_query(request, expected_response="fingerprint")
        return response.fingerprint.fingerprint

    def electrum_encryption_key(self, keypath: Sequence[int]) -> str:
        """
        This call fetches the xpub used for the electrum wallet encryption
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.electrum_encryption_key.CopyFrom(
            keystore.ElectrumEncryptionKeyRequest(keypath=keypath)
        )
        return self._msg_query(request).electrum_encryption_key.key

    def enable_mnemonic_passphrase(self) -> None:
        """
        Enable the bip39 passphrase.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.set_mnemonic_passphrase_enabled.enabled = True
        self._msg_query(request, expected_response="success")

    def disable_mnemonic_passphrase(self) -> None:
        """
        Disable the bip39 passphrase.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.set_mnemonic_passphrase_enabled.enabled = False
        self._msg_query(request, expected_response="success")

    def _eth_msg_query(
        self, eth_request: eth.ETHRequest, expected_response: Optional[str] = None
    ) -> eth.ETHResponse:
        """
        Same as _msg_query, but one nesting deeper for ethereum messages.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.eth.CopyFrom(eth_request)
        eth_response = self._msg_query(request, expected_response="eth").eth
        if (
            expected_response is not None
            and eth_response.WhichOneof("response") != expected_response
        ):
            raise Exception(
                "Unexpected response: {}, expected: {}".format(
                    eth_response.WhichOneof("response"), expected_response
                )
            )
        return eth_response

    def _eth_coin(self, chain_id: int) -> "eth.ETHCoin.V":
        """Returns the deprecated `coin` enum value for a given chain_id. Only ETH, Ropsten and Rinkeby are converted, as these were the only supported networks up to v9.10.0. With v9.10.0, the chain ID is passed directly, and the `coin` field is ignored."""
        if self.version < semver.VersionInfo(9, 10, 0):
            return {
                1: eth.ETHCoin.ETH,
                3: eth.ETHCoin.RopstenETH,
                4: eth.ETHCoin.RinkebyETH,
            }[chain_id]
        return eth.ETHCoin.ETH

    def eth_pub(
        self,
        keypath: Sequence[int],
        chain_id: int = 1,
        output_type: "eth.ETHPubRequest.OutputType.V" = eth.ETHPubRequest.ADDRESS,
        display: bool = True,
        contract_address: bytes = b"",
    ) -> str:
        """
        keypath is a list of child derivation numbers.
        e.g. m/44'/60'/0'/0/5 corresponds to [44+HARDENED, 60+HARDENED, 0+HARDENED, 0, 5].
        """
        # pylint: disable=no-member
        request = eth.ETHRequest()
        request.pub.CopyFrom(
            eth.ETHPubRequest(
                coin=self._eth_coin(chain_id),
                chain_id=chain_id,
                keypath=keypath,
                output_type=output_type,
                display=display,
                contract_address=contract_address,
            )
        )
        return self._eth_msg_query(request, expected_response="pub").pub.pub

    def eth_sign(self, transaction: bytes, keypath: Sequence[int], chain_id: int = 1) -> bytes:
        """
        transaction should be given as a full rlp encoded eth transaction.
        """
        nonce, gas_price, gas_limit, recipient, value, data, _, _, _ = rlp.decode(transaction)
        request = eth.ETHRequest()
        # pylint: disable=no-member
        request.sign.CopyFrom(
            eth.ETHSignRequest(
                coin=self._eth_coin(chain_id),
                chain_id=chain_id,
                keypath=keypath,
                nonce=nonce,
                gas_price=gas_price,
                gas_limit=gas_limit,
                recipient=recipient,
                value=value,
                data=data,
            )
        )

        supports_antiklepto = self.version >= semver.VersionInfo(9, 5, 0)
        if supports_antiklepto:
            host_nonce = os.urandom(32)

            request.sign.host_nonce_commitment.commitment = antiklepto_host_commit(host_nonce)
            signer_commitment = self._eth_msg_query(
                request, expected_response="antiklepto_signer_commitment"
            ).antiklepto_signer_commitment.commitment

            request = eth.ETHRequest()
            request.antiklepto_signature.CopyFrom(
                antiklepto.AntiKleptoSignatureRequest(host_nonce=host_nonce)
            )

            signature = self._eth_msg_query(request, expected_response="sign").sign.signature
            antiklepto_verify(host_nonce, signer_commitment, signature[:64])

            if self.debug:
                print("Antiklepto nonce verification PASSED")

            return signature

        return self._eth_msg_query(request, expected_response="sign").sign.signature

    def eth_sign_msg(self, msg: bytes, keypath: Sequence[int], chain_id: int = 1) -> bytes:
        """
        Signs message, the msg will be prefixed with "\x19Ethereum message\n" + len(msg) in the
        hardware. 27 is added to the recID to denote an uncompressed pubkey.
        """

        def format_as_uncompressed(sig: bytes) -> bytes:
            # 27 is the magic constant to add to the recoverable ID to denote an uncompressed
            # pubkey.
            modified_signature = list(sig)
            modified_signature[64] += 27
            return bytes(modified_signature)

        request = eth.ETHRequest()
        # pylint: disable=no-member
        request.sign_msg.CopyFrom(
            eth.ETHSignMessageRequest(
                coin=self._eth_coin(chain_id), chain_id=chain_id, keypath=keypath, msg=msg
            )
        )

        supports_antiklepto = self.version >= semver.VersionInfo(9, 5, 0)
        if supports_antiklepto:
            host_nonce = os.urandom(32)

            request.sign_msg.host_nonce_commitment.commitment = antiklepto_host_commit(host_nonce)
            signer_commitment = self._eth_msg_query(
                request, expected_response="antiklepto_signer_commitment"
            ).antiklepto_signer_commitment.commitment

            request = eth.ETHRequest()
            request.antiklepto_signature.CopyFrom(
                antiklepto.AntiKleptoSignatureRequest(host_nonce=host_nonce)
            )

            signature = self._eth_msg_query(request, expected_response="sign").sign.signature
            antiklepto_verify(host_nonce, signer_commitment, signature[:64])

            if self.debug:
                print("Antiklepto nonce verification PASSED")

            return format_as_uncompressed(signature)

        signature = self._eth_msg_query(request, expected_response="sign").sign.signature
        return format_as_uncompressed(signature)

    def eth_sign_typed_msg(
        self, keypath: Sequence[int], msg: Dict[str, Any], chain_id: int = 1
    ) -> bytes:
        """
        Sign a EIP-712 typed message.
        """
        # pylint: disable=too-many-statements

        self._require_atleast(semver.VersionInfo(9, 12, 0))

        def format_as_uncompressed(sig: bytes) -> bytes:
            # 27 is the magic constant to add to the recoverable ID to denote an uncompressed
            # pubkey.
            modified_signature = list(sig)
            modified_signature[64] += 27
            return bytes(modified_signature)

        request = eth.ETHRequest()

        host_nonce = os.urandom(32)

        def to_type(typ: str) -> eth.ETHSignTypedMessageRequest.MemberType:
            # pylint: disable=too-many-return-statements, no-member

            if typ.endswith("]"):
                rest, size = typ[:-1].rsplit("[", 1)
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.ARRAY,
                    size=int(size) if size else 0,
                    array_type=to_type(rest),
                )
            if typ.startswith("bytes"):
                size = typ[len("bytes") :]
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.BYTES,
                    size=int(size) if size else 0,
                )
            if typ.startswith("uint"):
                size = typ[len("uint") :]
                assert size and int(size) % 8 == 0
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.UINT,
                    size=int(size) // 8,
                )
            if typ.startswith("int"):
                size = typ[len("int") :]
                assert size and int(size) % 8 == 0
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.INT,
                    size=int(size) // 8,
                )
            if typ == "bool":
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.BOOL,
                )
            if typ == "address":
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.ADDRESS,
                )
            if typ == "string":
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.STRING,
                )
            if typ in msg["types"]:
                return eth.ETHSignTypedMessageRequest.MemberType(
                    type=eth.ETHSignTypedMessageRequest.DataType.STRUCT,
                    struct_name=typ,
                )
            raise ValueError("Unrecognized type: {}".format(typ))

        def get_value(
            root_object: "eth.ETHTypedMessageValueResponse.RootObject.V", path: Sequence[int]
        ) -> bytes:
            # pylint: disable=too-many-return-statements, too-many-branches, no-member

            if root_object == eth.ETHTypedMessageValueResponse.RootObject.DOMAIN:
                value = msg["domain"]
                typ = to_type("EIP712Domain")
            elif root_object == eth.ETHTypedMessageValueResponse.RootObject.MESSAGE:
                value = msg["message"]
                typ = to_type(msg["primaryType"])
            else:
                raise ValueError("Unknown root object: {}".format(root_object))

            for element in path:
                if typ.type == eth.ETHSignTypedMessageRequest.DataType.STRUCT:
                    struct_member = msg["types"][typ.struct_name][element]
                    value = value[struct_member["name"]]
                    typ = to_type(struct_member["type"])
                elif typ.type == eth.ETHSignTypedMessageRequest.DataType.ARRAY:
                    value = value[element]
                    typ = typ.array_type
                else:
                    raise ValueError("Path element does not point to struct or array")

            if typ.type == eth.ETHSignTypedMessageRequest.DataType.BYTES:
                if isinstance(value, str):
                    if value == "":
                        return b""
                    if value[:2].lower() == "0x":
                        return bytes.fromhex(value[2:])
                assert isinstance(value, bytes)
                return value
            if typ.type == eth.ETHSignTypedMessageRequest.DataType.UINT:
                if isinstance(value, str):
                    value = int(value)
                assert isinstance(value, int)
                return value.to_bytes(typ.size, "big")
            if typ.type == eth.ETHSignTypedMessageRequest.DataType.INT:
                if isinstance(value, str):
                    value = int(value)
                assert isinstance(value, int)
                return value.to_bytes(typ.size, "big", signed=True)
            if typ.type == eth.ETHSignTypedMessageRequest.DataType.BOOL:
                return bytes([value])
            if typ.type == eth.ETHSignTypedMessageRequest.DataType.ADDRESS:
                assert isinstance(value, str)
                return value.encode("ascii")
            if typ.type == eth.ETHSignTypedMessageRequest.DataType.STRING:
                assert isinstance(value, str)
                return value.encode("ascii")
            if typ.type == eth.ETHSignTypedMessageRequest.DataType.ARRAY:
                return len(value).to_bytes(4, "big")
            raise ValueError("Unexpected value query at path: {}. Type={}".format(path, typ))

        # pylint: disable=no-member
        request.sign_typed_msg.CopyFrom(
            eth.ETHSignTypedMessageRequest(
                chain_id=chain_id,
                keypath=keypath,
                types=[
                    eth.ETHSignTypedMessageRequest.StructType(
                        name=key,
                        members=[
                            eth.ETHSignTypedMessageRequest.Member(
                                name=member["name"],
                                type=to_type(member["type"]),
                            )
                            for member in val
                        ],
                    )
                    for key, val in msg["types"].items()
                ],
                primary_type=msg["primaryType"],
                host_nonce_commitment=antiklepto.AntiKleptoHostNonceCommitment(
                    commitment=antiklepto_host_commit(host_nonce),
                ),
            )
        )

        response = self._eth_msg_query(request)
        while response.WhichOneof("response") == "typed_msg_value":
            response = self._eth_msg_query(
                eth.ETHRequest(
                    typed_msg_value=eth.ETHTypedMessageValueRequest(
                        value=get_value(
                            response.typed_msg_value.root_object, response.typed_msg_value.path
                        ),
                    ),
                )
            )

        assert response.WhichOneof("response") == "antiklepto_signer_commitment"
        signer_commitment = response.antiklepto_signer_commitment.commitment

        request = eth.ETHRequest()
        request.antiklepto_signature.CopyFrom(
            antiklepto.AntiKleptoSignatureRequest(host_nonce=host_nonce)
        )

        signature = self._eth_msg_query(request, expected_response="sign").sign.signature
        antiklepto_verify(host_nonce, signer_commitment, signature[:64])

        if self.debug:
            print("Antiklepto nonce verification PASSED")

        return format_as_uncompressed(signature)

    def reset(self) -> bool:
        """
        Factory reset the device. Returns True on success.
        """
        request = hww.Request()
        # pylint: disable=no-member
        request.reset.CopyFrom(bitbox02_system.ResetRequest())
        try:
            self._msg_query(request)
        except OSError:
            # In case of reboot we can't read the response.
            return True
        except Bitbox02Exception as err:
            if err.code == ERR_GENERIC:
                return False
            raise
        return True

    def restore_from_mnemonic(self) -> None:
        """
        Restore from mnemonic. Raises a Bitbox02Exception on failure.
        """
        request = hww.Request()
        # pylint: disable=no-member
        request.restore_from_mnemonic.CopyFrom(
            mnemonic.RestoreFromMnemonicRequest(
                timestamp=int(time.time()), timezone_offset=time.localtime().tm_gmtoff
            )
        )
        self._msg_query(request)

    def _cardano_msg_query(
        self, cardano_request: cardano.CardanoRequest, expected_response: Optional[str] = None
    ) -> cardano.CardanoResponse:
        """
        Same as _msg_query, but one nesting deeper for cardano messages.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.cardano.CopyFrom(cardano_request)
        cardano_response = self._msg_query(request, expected_response="cardano").cardano
        if (
            expected_response is not None
            and cardano_response.WhichOneof("response") != expected_response
        ):
            raise Exception(
                "Unexpected response: {}, expected: {}".format(
                    cardano_response.WhichOneof("response"), expected_response
                )
            )
        return cardano_response

    def cardano_xpubs(self, keypaths: Sequence[Sequence[int]]) -> Sequence[bytes]:
        request = cardano.CardanoRequest(
            xpubs=cardano.CardanoXpubsRequest(
                keypaths=[common.Keypath(keypath=keypath) for keypath in keypaths]
            )
        )
        return self._cardano_msg_query(request, expected_response="xpubs").xpubs.xpubs

    def cardano_address(self, address: cardano.CardanoAddressRequest) -> str:
        # pylint: disable=no-member

        request = cardano.CardanoRequest(address=address)
        return self._cardano_msg_query(request, expected_response="pub").pub.pub

    def cardano_sign_transaction(
        self, transaction: cardano.CardanoSignTransactionRequest
    ) -> cardano.CardanoSignTransactionResponse:
        request = cardano.CardanoRequest(sign_transaction=transaction)
        return self._cardano_msg_query(
            request, expected_response="sign_transaction"
        ).sign_transaction



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: antiklepto.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61ntiklepto.proto\x12\x14shiftcrypto.bitbox02\"3\n\x1d\x41ntiKleptoHostNonceCommitment\x12\x12\n\ncommitment\x18\x01 \x01(\x0c\"0\n\x1a\x41ntiKleptoSignerCommitment\x12\x12\n\ncommitment\x18\x01 \x01(\x0c\"0\n\x1a\x41ntiKleptoSignatureRequest\x12\x12\n\nhost_nonce\x18\x01 \x01(\x0c\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'antiklepto_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ANTIKLEPTOHOSTNONCECOMMITMENT._serialized_start=42
  _ANTIKLEPTOHOSTNONCECOMMITMENT._serialized_end=93
  _ANTIKLEPTOSIGNERCOMMITMENT._serialized_start=95
  _ANTIKLEPTOSIGNERCOMMITMENT._serialized_end=143
  _ANTIKLEPTOSIGNATUREREQUEST._serialized_start=145
  _ANTIKLEPTOSIGNATUREREQUEST._serialized_end=193
# @@protoc_insertion_point(module_scope)



"""
Original version: https://github.com/bitcoin-core/HWI


MIT License

Copyright (c) 2017 Andrew Chow

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


Errors and Error Codes
**********************

HWI has several possible Exceptions with corresponding error codes.

:class:`~hwilib.hwwclient.HardwareWalletClient` functions and :mod:`~hwilib.commands` functions will generally raise an exception that is a subclass of :class:`HWWError`.
The HWI command line tool will convert these exceptions into a dictionary containing the error message and error code.
These look like ``{"error": "<msg>", "code": <code>}``.
"""

from typing import Any, Dict, Iterator, Optional
from contextlib import contextmanager

# Error codes
NO_DEVICE_TYPE = -1 #: Device type was not specified
MISSING_ARGUMENTS = -2 #: Arguments are missing
DEVICE_CONN_ERROR = -3 #: Error connecting to the device
UNKNWON_DEVICE_TYPE = -4 #: Device type is unknown
INVALID_TX = -5 #: Transaction is invalid
NO_PASSWORD = -6 #: No password provided, but one is needed
BAD_ARGUMENT = -7 #: Bad, malformed, or conflicting argument was provided
NOT_IMPLEMENTED = -8 #: Function is not implemented
UNAVAILABLE_ACTION = -9 #: Function is not available for this device
DEVICE_ALREADY_INIT = -10 #: Device is already initialized
DEVICE_ALREADY_UNLOCKED = -11 #: Device is already unlocked
DEVICE_NOT_READY = -12 #: Device is not ready
UNKNOWN_ERROR = -13 #: An unknown error occurred
ACTION_CANCELED = -14 #: Action was canceled by the user
DEVICE_BUSY = -15 #: Device is busy
NEED_TO_BE_ROOT = -16 #: User needs to be root to perform action
HELP_TEXT = -17 #: Help text was requested by the user
DEVICE_NOT_INITIALIZED = -18 #: Device is not initialized

# Exceptions
class HWWError(Exception):
    """
    Generic exception type produced by HWI
    Subclassed by specific Errors to have Exceptions that have specific error codes.

    Contains a message and error code.
    """

    def __init__(self, msg: str, code: int) -> None:
        """
        Create an exception with the message and error code

        :param msg: The error message
        :param code: The error code
        """
        Exception.__init__(self)
        self.code = code
        self.msg = msg

    def get_code(self) -> int:
        """
        Get the error code for this Error

        :return: The error code
        """
        return self.code

    def get_msg(self) -> str:
        """
        Get the error message for this Error

        :return: The error message
        """
        return self.msg

    def __str__(self) -> str:
        return self.msg

class NoPasswordError(HWWError):
    """
    :class:`HWWError` for :data:`NO_PASSWORD`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, NO_PASSWORD)

class UnavailableActionError(HWWError):
    """
    :class:`HWWError` for :data:`UNAVAILABLE_ACTION`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, UNAVAILABLE_ACTION)

class DeviceAlreadyInitError(HWWError):
    """
    :class:`HWWError` for :data:`DEVICE_ALREADY_INIT`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, DEVICE_ALREADY_INIT)

class DeviceNotReadyError(HWWError):
    """
    :class:`HWWError` for :data:`DEVICE_NOT_READY`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, DEVICE_NOT_READY)

class DeviceAlreadyUnlockedError(HWWError):
    """
    :class:`HWWError` for :data:`DEVICE_ALREADY_UNLOCKED`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, DEVICE_ALREADY_UNLOCKED)

class UnknownDeviceError(HWWError):
    """
    :class:`HWWError` for :data:`DEVICE_TYPE`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, UNKNWON_DEVICE_TYPE)

class NotImplementedError(HWWError):
    """
    :class:`HWWError` for :data:`NOT_IMPLEMENTED`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, NOT_IMPLEMENTED)

class PSBTSerializationError(HWWError):
    """
    :class:`HWWError` for :data:`INVALID_TX`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, INVALID_TX)

class BadArgumentError(HWWError):
    """
    :class:`HWWError` for :data:`BAD_ARGUMENT`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, BAD_ARGUMENT)

class DeviceFailureError(HWWError):
    """
    :class:`HWWError` for :data:`UNKNOWN_ERROR`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, UNKNOWN_ERROR)

class ActionCanceledError(HWWError):
    """
    :class:`HWWError` for :data:`ACTION_CANCELED`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, ACTION_CANCELED)

class DeviceConnectionError(HWWError):
    """
    :class:`HWWError` for :data:`DEVICE_CONN_ERROR`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, DEVICE_CONN_ERROR)

class DeviceBusyError(HWWError):
    """
    :class:`HWWError` for :data:`DEVICE_BUSY`
    """

    def __init__(self, msg: str):
        """
        :param msg: The error message
        """
        HWWError.__init__(self, msg, DEVICE_BUSY)

class NeedsRootError(HWWError):
    def __init__(self, msg: str):
        HWWError.__init__(self, msg, NEED_TO_BE_ROOT)

@contextmanager
def handle_errors(
    msg: Optional[str] = None,
    result: Optional[Dict[str, Any]] = None,
    code: int = UNKNOWN_ERROR,
    debug: bool = False,
) -> Iterator[None]:
    """
    Context manager to catch all Exceptions and HWWErrors to return them as dictionaries containing the error message and code.

    :param msg: Error message prefix. Attached to the beginning of each error message
    :param result: The dictionary to put the resulting error in
    :param code: The default error code to use for Exceptions
    :param debug: Whether to also print out the traceback for debugging purposes
    """
    if result is None:
        result = {}

    if msg is None:
        msg = ""
    else:
        msg = msg + " "

    try:
        yield

    except HWWError as e:
        result['error'] = msg + e.get_msg()
        result['code'] = e.get_code()
    except Exception as e:
        result['error'] = msg + str(e)
        result['code'] = code
        if debug:
            import traceback
            traceback.print_exc()
    return result


common_err_msgs = {
    "enumerate": "Could not open client or get fingerprint information:"
}



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'signmessagedialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_SignMessageDialog(object):
    def setupUi(self, SignMessageDialog):
        if not SignMessageDialog.objectName():
            SignMessageDialog.setObjectName(u"SignMessageDialog")
        SignMessageDialog.resize(600, 260)
        self.gridLayout = QGridLayout(SignMessageDialog)
        self.gridLayout.setSpacing(10)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(10, 10, 10, 10)
        self.msg_textedit = QPlainTextEdit(SignMessageDialog)
        self.msg_textedit.setObjectName(u"msg_textedit")
        self.msg_textedit.setMinimumSize(QSize(0, 50))
        self.msg_textedit.setMaximumSize(QSize(16777215, 100))

        self.gridLayout.addWidget(self.msg_textedit, 0, 1, 1, 2)

        self.path_lineedit = QLineEdit(SignMessageDialog)
        self.path_lineedit.setObjectName(u"path_lineedit")
        self.path_lineedit.setMinimumSize(QSize(300, 30))
        self.path_lineedit.setMaximumSize(QSize(16777215, 30))

        self.gridLayout.addWidget(self.path_lineedit, 1, 1, 1, 1)

        self.label_3 = QLabel(SignMessageDialog)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)

        self.label = QLabel(SignMessageDialog)
        self.label.setObjectName(u"label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.signmsg_button = QPushButton(SignMessageDialog)
        self.signmsg_button.setObjectName(u"signmsg_button")
        self.signmsg_button.setMaximumSize(QSize(150, 16777215))
        self.signmsg_button.setAutoDefault(False)

        self.gridLayout.addWidget(self.signmsg_button, 1, 2, 1, 1)

        self.sig_textedit = QPlainTextEdit(SignMessageDialog)
        self.sig_textedit.setObjectName(u"sig_textedit")
        self.sig_textedit.setMinimumSize(QSize(0, 30))
        self.sig_textedit.setMaximumSize(QSize(16777215, 50))
        self.sig_textedit.setTextInteractionFlags(Qt.TextSelectableByKeyboard|Qt.TextSelectableByMouse)

        self.gridLayout.addWidget(self.sig_textedit, 2, 1, 1, 2)

        self.label_2 = QLabel(SignMessageDialog)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)

        self.buttonBox = QDialogButtonBox(SignMessageDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setMinimumSize(QSize(80, 0))
        self.buttonBox.setMaximumSize(QSize(80, 16777215))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Close)

        self.gridLayout.addWidget(self.buttonBox, 3, 2, 1, 1, Qt.AlignRight)


        self.retranslateUi(SignMessageDialog)
        self.buttonBox.accepted.connect(SignMessageDialog.accept)
        self.buttonBox.rejected.connect(SignMessageDialog.reject)

        self.signmsg_button.setDefault(True)


        QMetaObject.connectSlotsByName(SignMessageDialog)
    # setupUi

    def retranslateUi(self, SignMessageDialog):
        SignMessageDialog.setWindowTitle(QCoreApplication.translate("SignMessageDialog", u"Dialog", None))
        self.label_3.setText(QCoreApplication.translate("SignMessageDialog", u"Signature", None))
        self.label.setText(QCoreApplication.translate("SignMessageDialog", u"Message", None))
        self.signmsg_button.setText(QCoreApplication.translate("SignMessageDialog", u"Sign Message", None))
        self.label_2.setText(QCoreApplication.translate("SignMessageDialog", u"Derivation Path", None))
    # retranslateUi




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hww.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2
from . import backup_commands_pb2 as backup__commands__pb2
from . import bitbox02_system_pb2 as bitbox02__system__pb2
from . import btc_pb2 as btc__pb2
from . import cardano_pb2 as cardano__pb2
from . import eth_pb2 as eth__pb2
from . import keystore_pb2 as keystore__pb2
from . import mnemonic_pb2 as mnemonic__pb2
from . import system_pb2 as system__pb2
from . import perform_attestation_pb2 as perform__attestation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\thww.proto\x12\x14shiftcrypto.bitbox02\x1a\x0c\x63ommon.proto\x1a\x15\x62\x61\x63kup_commands.proto\x1a\x15\x62itbox02_system.proto\x1a\tbtc.proto\x1a\rcardano.proto\x1a\teth.proto\x1a\x0ekeystore.proto\x1a\x0emnemonic.proto\x1a\x0csystem.proto\x1a\x19perform_attestation.proto\"&\n\x05\x45rror\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\"\t\n\x07Success\"\xc8\r\n\x07Request\x12\x41\n\x0b\x64\x65vice_name\x18\x02 \x01(\x0b\x32*.shiftcrypto.bitbox02.SetDeviceNameRequestH\x00\x12I\n\x0f\x64\x65vice_language\x18\x03 \x01(\x0b\x32..shiftcrypto.bitbox02.SetDeviceLanguageRequestH\x00\x12>\n\x0b\x64\x65vice_info\x18\x04 \x01(\x0b\x32\'.shiftcrypto.bitbox02.DeviceInfoRequestH\x00\x12@\n\x0cset_password\x18\x05 \x01(\x0b\x32(.shiftcrypto.bitbox02.SetPasswordRequestH\x00\x12\x42\n\rcreate_backup\x18\x06 \x01(\x0b\x32).shiftcrypto.bitbox02.CreateBackupRequestH\x00\x12\x42\n\rshow_mnemonic\x18\x07 \x01(\x0b\x32).shiftcrypto.bitbox02.ShowMnemonicRequestH\x00\x12\x36\n\x07\x62tc_pub\x18\x08 \x01(\x0b\x32#.shiftcrypto.bitbox02.BTCPubRequestH\x00\x12\x41\n\rbtc_sign_init\x18\t \x01(\x0b\x32(.shiftcrypto.bitbox02.BTCSignInitRequestH\x00\x12\x43\n\x0e\x62tc_sign_input\x18\n \x01(\x0b\x32).shiftcrypto.bitbox02.BTCSignInputRequestH\x00\x12\x45\n\x0f\x62tc_sign_output\x18\x0b \x01(\x0b\x32*.shiftcrypto.bitbox02.BTCSignOutputRequestH\x00\x12O\n\x14insert_remove_sdcard\x18\x0c \x01(\x0b\x32/.shiftcrypto.bitbox02.InsertRemoveSDCardRequestH\x00\x12@\n\x0c\x63heck_sdcard\x18\r \x01(\x0b\x32(.shiftcrypto.bitbox02.CheckSDCardRequestH\x00\x12\x64\n\x1fset_mnemonic_passphrase_enabled\x18\x0e \x01(\x0b\x32\x39.shiftcrypto.bitbox02.SetMnemonicPassphraseEnabledRequestH\x00\x12@\n\x0clist_backups\x18\x0f \x01(\x0b\x32(.shiftcrypto.bitbox02.ListBackupsRequestH\x00\x12\x44\n\x0erestore_backup\x18\x10 \x01(\x0b\x32*.shiftcrypto.bitbox02.RestoreBackupRequestH\x00\x12N\n\x13perform_attestation\x18\x11 \x01(\x0b\x32/.shiftcrypto.bitbox02.PerformAttestationRequestH\x00\x12\x35\n\x06reboot\x18\x12 \x01(\x0b\x32#.shiftcrypto.bitbox02.RebootRequestH\x00\x12@\n\x0c\x63heck_backup\x18\x13 \x01(\x0b\x32(.shiftcrypto.bitbox02.CheckBackupRequestH\x00\x12/\n\x03\x65th\x18\x14 \x01(\x0b\x32 .shiftcrypto.bitbox02.ETHRequestH\x00\x12\x33\n\x05reset\x18\x15 \x01(\x0b\x32\".shiftcrypto.bitbox02.ResetRequestH\x00\x12Q\n\x15restore_from_mnemonic\x18\x16 \x01(\x0b\x32\x30.shiftcrypto.bitbox02.RestoreFromMnemonicRequestH\x00\x12\x43\n\x0b\x66ingerprint\x18\x18 \x01(\x0b\x32,.shiftcrypto.bitbox02.RootFingerprintRequestH\x00\x12/\n\x03\x62tc\x18\x19 \x01(\x0b\x32 .shiftcrypto.bitbox02.BTCRequestH\x00\x12U\n\x17\x65lectrum_encryption_key\x18\x1a \x01(\x0b\x32\x32.shiftcrypto.bitbox02.ElectrumEncryptionKeyRequestH\x00\x12\x37\n\x07\x63\x61rdano\x18\x1b \x01(\x0b\x32$.shiftcrypto.bitbox02.CardanoRequestH\x00\x42\t\n\x07requestJ\x04\x08\x01\x10\x02J\x04\x08\x17\x10\x18\"\x89\x07\n\x08Response\x12\x30\n\x07success\x18\x01 \x01(\x0b\x32\x1d.shiftcrypto.bitbox02.SuccessH\x00\x12,\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x1b.shiftcrypto.bitbox02.ErrorH\x00\x12?\n\x0b\x64\x65vice_info\x18\x04 \x01(\x0b\x32(.shiftcrypto.bitbox02.DeviceInfoResponseH\x00\x12\x30\n\x03pub\x18\x05 \x01(\x0b\x32!.shiftcrypto.bitbox02.PubResponseH\x00\x12\x42\n\rbtc_sign_next\x18\x06 \x01(\x0b\x32).shiftcrypto.bitbox02.BTCSignNextResponseH\x00\x12\x41\n\x0clist_backups\x18\x07 \x01(\x0b\x32).shiftcrypto.bitbox02.ListBackupsResponseH\x00\x12\x41\n\x0c\x63heck_backup\x18\x08 \x01(\x0b\x32).shiftcrypto.bitbox02.CheckBackupResponseH\x00\x12O\n\x13perform_attestation\x18\t \x01(\x0b\x32\x30.shiftcrypto.bitbox02.PerformAttestationResponseH\x00\x12\x41\n\x0c\x63heck_sdcard\x18\n \x01(\x0b\x32).shiftcrypto.bitbox02.CheckSDCardResponseH\x00\x12\x30\n\x03\x65th\x18\x0b \x01(\x0b\x32!.shiftcrypto.bitbox02.ETHResponseH\x00\x12\x44\n\x0b\x66ingerprint\x18\x0c \x01(\x0b\x32-.shiftcrypto.bitbox02.RootFingerprintResponseH\x00\x12\x30\n\x03\x62tc\x18\r \x01(\x0b\x32!.shiftcrypto.bitbox02.BTCResponseH\x00\x12V\n\x17\x65lectrum_encryption_key\x18\x0e \x01(\x0b\x32\x33.shiftcrypto.bitbox02.ElectrumEncryptionKeyResponseH\x00\x12\x38\n\x07\x63\x61rdano\x18\x0f \x01(\x0b\x32%.shiftcrypto.bitbox02.CardanoResponseH\x00\x42\n\n\x08responseJ\x04\x08\x03\x10\x04\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'hww_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ERROR._serialized_start=205
  _ERROR._serialized_end=243
  _SUCCESS._serialized_start=245
  _SUCCESS._serialized_end=254
  _REQUEST._serialized_start=257
  _REQUEST._serialized_end=1993
  _RESPONSE._serialized_start=1996
  _RESPONSE._serialized_end=2901
# @@protoc_insertion_point(module_scope)



from typing import Tuple, List, Mapping, Optional, Union
import base64
from io import BytesIO, BufferedReader

from .command_builder import BitcoinCommandBuilder, BitcoinInsType
from ...common import Chain
from .client_command import ClientCommandInterpreter
from .client_base import Client, TransportClient
from .client_legacy import LegacyClient
from .exception import DeviceException, NotSupportedError
from .merkle import get_merkleized_map_commitment
from .wallet import WalletPolicy, WalletType
from ...psbt import PSBT
from ..._serialize import deser_string


def parse_stream_to_map(f: BufferedReader) -> Mapping[bytes, bytes]:
    result = {}
    while True:
        try:
            key = deser_string(f)
        except Exception:
            break

        # Check for separator
        if len(key) == 0:
            break

        value = deser_string(f)

        result[key] = value
    return result


def read_uint(buf: BytesIO,
              bit_len: int,
              byteorder: str = 'little') -> int:
    assert byteorder in ['little', 'big']

    size: int = bit_len // 8
    b: bytes = buf.read(size)

    if len(b) < size:
        raise ValueError(f"Can't read u{bit_len} in buffer!")

    return int.from_bytes(b, byteorder)


def read_varint(buf: BytesIO,
                prefix: Optional[bytes] = None) -> int:
    b: bytes = prefix if prefix else buf.read(1)

    if not b:
        raise ValueError(f"Can't read prefix: '{b}'!")

    n: int = {b"\xfd": 2, b"\xfe": 4, b"\xff": 8}.get(b, 1)  # default to 1

    b = buf.read(n) if n > 1 else b

    if len(b) != n:
        raise ValueError("Can't read varint!")

    return int.from_bytes(b, byteorder="little")


class NewClient(Client):
    # internal use for testing: if set to True, sign_psbt will not clone the psbt before converting to psbt version 2
    _no_clone_psbt: bool = False

    def __init__(self, comm_client: TransportClient, chain: Chain = Chain.MAIN) -> None:
        super().__init__(comm_client, chain)
        self.builder = BitcoinCommandBuilder()

    # Modifies the behavior of the base method by taking care of SW_INTERRUPTED_EXECUTION responses
    def _make_request(
        self, apdu: dict, client_intepreter: ClientCommandInterpreter = None
    ) -> Tuple[int, bytes]:
        sw, response = self._apdu_exchange(apdu)

        while sw == 0xE000:
            if not client_intepreter:
                raise RuntimeError("Unexpected SW_INTERRUPTED_EXECUTION received.")

            command_response = client_intepreter.execute(response)
            sw, response = self._apdu_exchange(
                self.builder.continue_interrupted(command_response)
            )

        return sw, response

    def get_extended_pubkey(self, path: str, display: bool = False) -> str:
        sw, response = self._make_request(self.builder.get_extended_pubkey(path, display))

        if sw != 0x9000:
            raise DeviceException(error_code=sw, ins=BitcoinInsType.GET_EXTENDED_PUBKEY)

        return response.decode()

    def register_wallet(self, wallet: WalletPolicy) -> Tuple[bytes, bytes]:
        if wallet.version not in [WalletType.WALLET_POLICY_V1, WalletType.WALLET_POLICY_V2]:
            raise ValueError("invalid wallet policy version")

        client_intepreter = ClientCommandInterpreter()
        client_intepreter.add_known_preimage(wallet.serialize())
        client_intepreter.add_known_list([k.encode() for k in wallet.keys_info])

        # necessary for version 1 of the protocol (available since version 2.1.0 of the app)
        client_intepreter.add_known_preimage(wallet.descriptor_template.encode())

        sw, response = self._make_request(
            self.builder.register_wallet(wallet), client_intepreter
        )

        if sw != 0x9000:
            raise DeviceException(error_code=sw, ins=BitcoinInsType.REGISTER_WALLET)

        if len(response) != 64:
            raise RuntimeError(f"Invalid response length: {len(response)}")

        wallet_id = response[0:32]
        wallet_hmac = response[32:64]

        return wallet_id, wallet_hmac

    def get_wallet_address(
        self,
        wallet: WalletPolicy,
        wallet_hmac: Optional[bytes],
        change: int,
        address_index: int,
        display: bool,
    ) -> str:

        if not isinstance(wallet, WalletPolicy) or wallet.version not in [WalletType.WALLET_POLICY_V1, WalletType.WALLET_POLICY_V2]:
            raise ValueError("wallet type must be WalletPolicy, with version either WALLET_POLICY_V1 or WALLET_POLICY_V2")

        if change != 0 and change != 1:
            raise ValueError("Invalid change")

        client_intepreter = ClientCommandInterpreter()
        client_intepreter.add_known_list([k.encode() for k in wallet.keys_info])
        client_intepreter.add_known_preimage(wallet.serialize())

        # necessary for version 1 of the protocol (available since version 2.1.0 of the app)
        client_intepreter.add_known_preimage(wallet.descriptor_template.encode())

        sw, response = self._make_request(
            self.builder.get_wallet_address(
                wallet, wallet_hmac, address_index, change, display
            ),
            client_intepreter,
        )

        if sw != 0x9000:
            raise DeviceException(error_code=sw, ins=BitcoinInsType.GET_WALLET_ADDRESS)

        return response.decode()

    def sign_psbt(self, psbt: PSBT, wallet: WalletPolicy, wallet_hmac: Optional[bytes]) -> List[Tuple[int, bytes, bytes]]:
        """Signs a PSBT using a registered wallet (or a standard wallet that does not need registration).

        Signature requires explicit approval from the user.

        Parameters
        ----------
        psbt : PSBT
            A PSBT of version 0 or 2, with all the necessary information to sign the inputs already filled in; what the
            required fields changes depending on the type of input.
            The non-witness UTXO must be present for both legacy and SegWit inputs, or the hardware wallet will reject
            signing. This is not required for Taproot inputs.

        wallet : WalletPolicy
            The registered wallet policy, or a standard wallet policy.

        wallet_hmac: Optional[bytes]
            For a registered wallet, the hmac obtained at wallet registration. `None` for a standard wallet policy.

        Returns
        -------
        List[Tuple[int, bytes, bytes]]
            A list of tuples returned by the hardware wallets, where each element is a tuple of:
            - an integer, the index of the input being signed;
            - a `bytes` array of length 33 (compressed ecdsa pubkey) or 32 (x-only BIP-0340 pubkey), the corresponding pubkey for this signature;
            - a `bytes` array with the signature.
        """
        assert psbt.version == 2
        psbt_v2 = psbt

        psbt_bytes = base64.b64decode(psbt_v2.serialize())
        f = BytesIO(psbt_bytes)

        # We parse the individual maps (global map, each input map, and each output map) from the psbt serialized as a
        # sequence of bytes, in order to produce the serialized Merkleized map commitments. Moreover, we prepare the
        # client interpreter to respond on queries on all the relevant Merkle trees and pre-images in the psbt.

        assert f.read(5) == b"psbt\xff"

        client_intepreter = ClientCommandInterpreter()
        client_intepreter.add_known_list([k.encode() for k in wallet.keys_info])
        client_intepreter.add_known_preimage(wallet.serialize())

        # necessary for version 1 of the protocol (available since version 2.1.0 of the app)
        client_intepreter.add_known_preimage(wallet.descriptor_template.encode())

        global_map: Mapping[bytes, bytes] = parse_stream_to_map(f)
        client_intepreter.add_known_mapping(global_map)

        input_maps: List[Mapping[bytes, bytes]] = []
        for _ in range(len(psbt_v2.inputs)):
            input_maps.append(parse_stream_to_map(f))
        for m in input_maps:
            client_intepreter.add_known_mapping(m)

        output_maps: List[Mapping[bytes, bytes]] = []
        for _ in range(len(psbt_v2.outputs)):
            output_maps.append(parse_stream_to_map(f))
        for m in output_maps:
            client_intepreter.add_known_mapping(m)

        # We also add the Merkle tree of the input (resp. output) map commitments as a known tree
        input_commitments = [get_merkleized_map_commitment(m_in) for m_in in input_maps]
        output_commitments = [get_merkleized_map_commitment(m_out) for m_out in output_maps]

        client_intepreter.add_known_list(input_commitments)
        client_intepreter.add_known_list(output_commitments)

        sw, _ = self._make_request(
            self.builder.sign_psbt(
                global_map, input_maps, output_maps, wallet, wallet_hmac
            ),
            client_intepreter,
        )

        if sw != 0x9000:
            raise DeviceException(error_code=sw, ins=BitcoinInsType.SIGN_PSBT)

        # parse results and return a structured version instead
        results = client_intepreter.yielded

        if any(len(x) <= 1 for x in results):
            raise RuntimeError("Invalid response")

        results_list: List[Tuple[int, bytes, bytes]] = []
        for res in results:
            res_buffer = BytesIO(res)
            input_index = read_varint(res_buffer)

            pubkey_len = read_uint(res_buffer, 8)
            pubkey = res_buffer.read(pubkey_len)

            signature = res_buffer.read()

            results_list.append((input_index, pubkey, signature))

        return results_list

    def get_master_fingerprint(self) -> bytes:

        sw, response = self._make_request(self.builder.get_master_fingerprint())

        if sw != 0x9000:
            raise DeviceException(error_code=sw, ins=BitcoinInsType.GET_EXTENDED_PUBKEY)

        return response

    def sign_message(self, message: Union[str, bytes], bip32_path: str) -> str:
        if isinstance(message, str):
            message_bytes = message.encode("utf-8")
        else:
            message_bytes = message

        chunks = [message_bytes[64 * i: 64 * i + 64] for i in range((len(message_bytes) + 63) // 64)]

        client_intepreter = ClientCommandInterpreter()
        client_intepreter.add_known_list(chunks)

        sw, response = self._make_request(self.builder.sign_message(message_bytes, bip32_path), client_intepreter)

        if sw != 0x9000:
            raise DeviceException(error_code=sw, ins=BitcoinInsType.GET_EXTENDED_PUBKEY)

        return base64.b64encode(response).decode('utf-8')


def createClient(comm_client: Optional[TransportClient] = None, chain: Chain = Chain.MAIN, debug: bool = False) -> Union[LegacyClient, NewClient]:
    if comm_client is None:
        comm_client = TransportClient("hid", debug=debug)

    base_client = Client(comm_client, chain)
    app_name, app_version, _ = base_client.get_version()

    if app_name not in ["Bitcoin", "Bitcoin Test", "Bitcoin Legacy", "Bitcoin Test Legacy", "app"]:
        raise NotSupportedError(0x6A82, None, "Ledger is not in either the Bitcoin or Bitcoin Testnet app")

    app_version_major, app_version_minor, _ = app_version.split(".", 2)

    # App versions using the legacy protocol:
    # - any 1.*.* version
    # - any 2.0.* version
    # - any version if the app name is "Bitcoin Legacy" or "Bitcoin Test Legacy"
    is_legacy = int(app_version_major) <= 1 or (int(app_version_major) == 2 and int(app_version_minor) == 0) or "Legacy" in app_name

    if is_legacy:
        return LegacyClient(comm_client, chain)
    else:
        return NewClient(comm_client, chain)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class CheckSDCardRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___CheckSDCardRequest = CheckSDCardRequest

class CheckSDCardResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    INSERTED_FIELD_NUMBER: builtins.int
    inserted: builtins.bool
    def __init__(self,
        *,
        inserted: builtins.bool = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["inserted",b"inserted"]) -> None: ...
global___CheckSDCardResponse = CheckSDCardResponse

class DeviceInfoRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___DeviceInfoRequest = DeviceInfoRequest

class DeviceInfoResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    INITIALIZED_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    MNEMONIC_PASSPHRASE_ENABLED_FIELD_NUMBER: builtins.int
    MONOTONIC_INCREMENTS_REMAINING_FIELD_NUMBER: builtins.int
    SECURECHIP_MODEL_FIELD_NUMBER: builtins.int
    name: typing.Text
    initialized: builtins.bool
    version: typing.Text
    mnemonic_passphrase_enabled: builtins.bool
    monotonic_increments_remaining: builtins.int
    securechip_model: typing.Text
    """From v9.6.0: "ATECC608A" or "ATECC608B"."""

    def __init__(self,
        *,
        name: typing.Text = ...,
        initialized: builtins.bool = ...,
        version: typing.Text = ...,
        mnemonic_passphrase_enabled: builtins.bool = ...,
        monotonic_increments_remaining: builtins.int = ...,
        securechip_model: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["initialized",b"initialized","mnemonic_passphrase_enabled",b"mnemonic_passphrase_enabled","monotonic_increments_remaining",b"monotonic_increments_remaining","name",b"name","securechip_model",b"securechip_model","version",b"version"]) -> None: ...
global___DeviceInfoResponse = DeviceInfoResponse

class InsertRemoveSDCardRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _SDCardAction:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _SDCardActionEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[InsertRemoveSDCardRequest._SDCardAction.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        REMOVE_CARD: InsertRemoveSDCardRequest._SDCardAction.ValueType  # 0
        INSERT_CARD: InsertRemoveSDCardRequest._SDCardAction.ValueType  # 1
    class SDCardAction(_SDCardAction, metaclass=_SDCardActionEnumTypeWrapper):
        pass

    REMOVE_CARD: InsertRemoveSDCardRequest.SDCardAction.ValueType  # 0
    INSERT_CARD: InsertRemoveSDCardRequest.SDCardAction.ValueType  # 1

    ACTION_FIELD_NUMBER: builtins.int
    action: global___InsertRemoveSDCardRequest.SDCardAction.ValueType
    def __init__(self,
        *,
        action: global___InsertRemoveSDCardRequest.SDCardAction.ValueType = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["action",b"action"]) -> None: ...
global___InsertRemoveSDCardRequest = InsertRemoveSDCardRequest

class ResetRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___ResetRequest = ResetRequest

class SetDeviceLanguageRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    LANGUAGE_FIELD_NUMBER: builtins.int
    language: typing.Text
    def __init__(self,
        *,
        language: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["language",b"language"]) -> None: ...
global___SetDeviceLanguageRequest = SetDeviceLanguageRequest

class SetDeviceNameRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    name: typing.Text
    def __init__(self,
        *,
        name: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name",b"name"]) -> None: ...
global___SetDeviceNameRequest = SetDeviceNameRequest

class SetPasswordRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ENTROPY_FIELD_NUMBER: builtins.int
    entropy: builtins.bytes
    def __init__(self,
        *,
        entropy: builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["entropy",b"entropy"]) -> None: ...
global___SetPasswordRequest = SetPasswordRequest



# Copyright 2019 Shift Cryptosecurity AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Interact with a BitBox02 bootloader. """

import struct
import typing
import io
import math
import hashlib

from ..communication import TransportLayer
from ..communication.devices import DeviceInfo

BOOTLOADER_CMD = 0x80 + 0x40 + 0x03
NUM_ROOT_KEYS = 3
NUM_SIGNING_KEYS = 3

MAX_FIRMWARE_SIZE = 884736  # 928kB - 64kB
CHUNK_SIZE = 4096
# max number of chunks that can be written when flashing the firmware
assert MAX_FIRMWARE_SIZE % CHUNK_SIZE == 0
FIRMWARE_CHUNKS = MAX_FIRMWARE_SIZE // CHUNK_SIZE

SIGDATA_MAGIC_STANDARD = struct.pack(">I", 0x653F362B)
SIGDATA_MAGIC_BTCONLY = struct.pack(">I", 0x11233B0B)
SIGDATA_MAGIC_BITBOXBASE_STANDARD = struct.pack(">I", 0xAB6BD345)

MAGIC_LEN = 4

VERSION_LEN = 4
SIGNING_PUBKEYS_DATA_LEN = VERSION_LEN + NUM_SIGNING_KEYS * 64 + NUM_ROOT_KEYS * 64
FIRMWARE_DATA_LEN = VERSION_LEN + NUM_SIGNING_KEYS * 64
SIGDATA_LEN = SIGNING_PUBKEYS_DATA_LEN + FIRMWARE_DATA_LEN


def parse_signed_firmware(firmware: bytes) -> typing.Tuple[bytes, bytes, bytes]:
    """
    Split raw firmware bytes into magic, sigdata and firmware
    """

    if len(firmware) < MAGIC_LEN + SIGDATA_LEN:
        raise ValueError("firmware too small")
    magic, firmware = firmware[:MAGIC_LEN], firmware[MAGIC_LEN:]
    if magic not in (
        SIGDATA_MAGIC_STANDARD,
        SIGDATA_MAGIC_BTCONLY,
        SIGDATA_MAGIC_BITBOXBASE_STANDARD,
    ):
        raise ValueError("invalid magic")

    sigdata, firmware = firmware[:SIGDATA_LEN], firmware[SIGDATA_LEN:]
    return magic, sigdata, firmware


class Bootloader:
    """
    One instance of a BitBox02 Bootloader, exposing the bootloader API.
    """

    def __init__(self, transport: TransportLayer, device_info: DeviceInfo):
        self._transport = transport
        self.expected_magic = {
            "bb02-bootloader": SIGDATA_MAGIC_STANDARD,
            "bb02btc-bootloader": SIGDATA_MAGIC_BTCONLY,
            "bitboxbase-bootloader": SIGDATA_MAGIC_BITBOXBASE_STANDARD,
        }.get(device_info["product_string"])
        assert self.expected_magic

    def _query(self, msg: bytes) -> bytes:
        cid = self._transport.generate_cid()
        response = self._transport.query(msg, BOOTLOADER_CMD, cid)
        if response[0] != msg[0]:
            raise Exception("bootloader api error, expected {}, got {}".format(msg[0], response[0]))
        if response[1] != 0:
            raise Exception("bootloader api error: code={}".format(response[1]))
        return response[2:]

    def versions(self) -> typing.Tuple[int, int]:
        """
        Returns (firmware version, signing pubkeys version).
        """
        response = self._query(b"v")
        firmware_v, signing_pubkeys_v = struct.unpack("<II", response[:8])
        return firmware_v, signing_pubkeys_v

    def get_hashes(
        self, display_firmware_hash: bool = False, display_signing_keydata_hash: bool = False
    ) -> typing.Tuple[bytes, bytes]:
        """
        Returns (firmare hash, signing keydata hash).
        If display is True, the hash is shown on the device screen.
        """
        response = self._query(
            b"h" + bytes([int(display_firmware_hash), int(display_signing_keydata_hash)])
        )
        firmware_hash, signing_keydata_hash = response[:32], response[32:64]
        return firmware_hash, signing_keydata_hash

    def show_firmware_hash_enabled(self) -> bool:
        """
        Returns whether the bootloader will automatically show the firmware hash on boot.
        """
        return bool(self._query(b"H\xFF")[0])

    def set_show_firmware_hash(self, enable: bool) -> None:
        """
        Enables/disables whether the bootloader will automatically show the firmware hash on boot.
        """
        self._query(b"H" + bytes([int(enable)]))

    def _erase(self, firmware_num_chunks: int) -> None:
        self._query(b"e" + bytes([firmware_num_chunks]))

    def _write_chunk(self, chunk_num: int, chunk: bytes) -> None:
        if len(chunk) != CHUNK_SIZE:
            raise ValueError("chunk must be 4kB")
        self._query(b"w" + bytes([chunk_num]) + chunk)

    def flash_unsigned_firmware(
        self,
        firmware: bytes,
        progress_callback: typing.Optional[typing.Callable[[float], None]] = None,
    ) -> None:
        """
        Flashes a firmware image onto the bootloader by invoking the erase and write chunk api
        calls. Expects the raw firmware without signatures, and does not flash the signatures.
        """
        if len(firmware) > FIRMWARE_CHUNKS * CHUNK_SIZE:
            raise ValueError("firmware too big")
        if progress_callback is not None:
            progress_callback(0)
        num_chunks = math.ceil(len(firmware) / CHUNK_SIZE)
        self._erase(num_chunks)
        stream = io.BytesIO(firmware)
        chunk_num = 0
        while True:
            chunk = stream.read(CHUNK_SIZE)
            if not chunk:
                break
            if len(chunk) < CHUNK_SIZE:
                chunk += b"\xff" * (CHUNK_SIZE - len(chunk))
            self._write_chunk(chunk_num, chunk)
            chunk_num += 1
            if progress_callback is not None:
                progress_callback(chunk_num / num_chunks)

    def flash_signed_firmware(
        self,
        firmware: bytes,
        progress_callback: typing.Optional[typing.Callable[[float], None]] = None,
    ) -> None:
        """
        Flashes a signed firmware image. The firmware itself is extracted and flashed, then the
        signatures are extracted and flashed.
        """

        magic, sigdata, firmware = parse_signed_firmware(firmware)
        if magic != self.expected_magic:
            raise ValueError("wrong firmware edition")
        self.flash_unsigned_firmware(firmware, progress_callback=progress_callback)
        self._query(b"s" + sigdata)

    def erase(self) -> None:
        """
        Erases the firmware from the device.
        """
        self._erase(0)

    def erased(self) -> bool:
        """
        Returns True if the the device contains no firmware.
        """
        # We check by comparing the device reported firmware hash.
        # If erased, the firmware is all '\xFF'.
        firmware_v, _ = self.versions()
        empty_firmware = struct.pack("<I", firmware_v) + b"\xFF" * MAX_FIRMWARE_SIZE
        empty_firmware_hash = hashlib.sha256(hashlib.sha256(empty_firmware).digest()).digest()
        reported_firmware_hash, _ = self.get_hashes()
        return empty_firmware_hash == reported_firmware_hash

    def reboot(self) -> None:
        self._transport.write(b"r", BOOTLOADER_CMD, self._transport.generate_cid())
        self._transport.close()

    def screen_rotate(self) -> None:
        self._query(b"f")

    def close(self) -> None:
        self._transport.close()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class PubResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    PUB_FIELD_NUMBER: builtins.int
    pub: typing.Text
    def __init__(self,
        *,
        pub: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["pub",b"pub"]) -> None: ...
global___PubResponse = PubResponse

class RootFingerprintRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___RootFingerprintRequest = RootFingerprintRequest

class RootFingerprintResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FINGERPRINT_FIELD_NUMBER: builtins.int
    fingerprint: builtins.bytes
    def __init__(self,
        *,
        fingerprint: builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["fingerprint",b"fingerprint"]) -> None: ...
global___RootFingerprintResponse = RootFingerprintResponse

class XPub(google.protobuf.message.Message):
    """See https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki.
    version field dropped as it will set dynamically based on the context (xpub, ypub, etc.).
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    DEPTH_FIELD_NUMBER: builtins.int
    PARENT_FINGERPRINT_FIELD_NUMBER: builtins.int
    CHILD_NUM_FIELD_NUMBER: builtins.int
    CHAIN_CODE_FIELD_NUMBER: builtins.int
    PUBLIC_KEY_FIELD_NUMBER: builtins.int
    depth: builtins.bytes
    parent_fingerprint: builtins.bytes
    child_num: builtins.int
    chain_code: builtins.bytes
    public_key: builtins.bytes
    def __init__(self,
        *,
        depth: builtins.bytes = ...,
        parent_fingerprint: builtins.bytes = ...,
        child_num: builtins.int = ...,
        chain_code: builtins.bytes = ...,
        public_key: builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["chain_code",b"chain_code","child_num",b"child_num","depth",b"depth","parent_fingerprint",b"parent_fingerprint","public_key",b"public_key"]) -> None: ...
global___XPub = XPub

class Keypath(google.protobuf.message.Message):
    """This message exists for use in oneof or repeated fields, where one can't inline `repeated uint32` due to protobuf rules."""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    KEYPATH_FIELD_NUMBER: builtins.int
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(self,
        *,
        keypath: typing.Optional[typing.Iterable[builtins.int]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["keypath",b"keypath"]) -> None: ...
global___Keypath = Keypath
